# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lattice route"""

import json
import uuid
from typing import Optional

from fastapi import APIRouter, HTTPException, Path, Query
from sqlalchemy.orm import Session

import covalent_ui.api.v1.database.config.db as db
from covalent_ui.api.v1.data_layer.lattice_dal import Lattices
from covalent_ui.api.v1.models.dispatch_model import SortDirection
from covalent_ui.api.v1.models.lattices_model import (
    LatticeDetailResponse,
    LatticeExecutorResponse,
    LatticeFileOutput,
    LatticeFileResponse,
    LatticeWorkflowExecutorResponse,
    SubLatticeDetailResponse,
    SubLatticeSortBy,
)
from covalent_ui.api.v1.utils.file_handle import FileHandler

routes: APIRouter = APIRouter()


@routes.get("/{dispatch_id}", response_model=LatticeDetailResponse)
def get_lattice_details(dispatch_id: uuid.UUID):
    """Get lattice details

    Args:
        dispatch_id: To fetch lattice data with the provided dispatch id

    Returns:
        Returns the lattice data with the dispatch id provided
    """

    with Session(db.engine) as session:
        lattice = Lattices(session)
        data = lattice.get_lattices_id(dispatch_id)
        if data is not None:
            handler = FileHandler(data["directory"])
            return LatticeDetailResponse(
                dispatch_id=data.dispatch_id,
                status=data.status,
                total_electrons=data.total_electrons,
                total_electrons_completed=data.total_electrons_completed,
                started_at=data.start_time,
                ended_at=data.end_time,
                directory=data.directory,
                description=handler.read_from_text(data.docstring_filename),
                runtime=data.runtime,
            )
        raise HTTPException(
            status_code=400,
            detail=[
                {
                    "loc": ["path", "dispatch_id"],
                    "msg": f"Dispatch ID {dispatch_id} does not exist",
                    "type": None,
                }
            ],
        )


@routes.get("/{dispatch_id}/details/{name}")
def get_lattice_files(dispatch_id: uuid.UUID, name: LatticeFileOutput):
    """Get lattice file data

    Args:
        dispatch_id: To fetch lattice data with the provided dispatch id
        name: To fetch specific file data for a lattice

    Returns:
        Returns the lattice file data with the dispatch id and file_module provided provided
    """
    with Session(db.engine) as session:
        lattice = Lattices(session)
        lattice_data = lattice.get_lattices_id_storage_file(dispatch_id)

        if lattice_data is None:
            raise HTTPException(
                status_code=400,
                detail=[
                    {
                        "loc": ["path", "dispatch_id"],
                        "msg": f"Dispatch ID {dispatch_id} does not exist",
                        "type": None,
                    }
                ],
            )

        handler = FileHandler(lattice_data["directory"])
        if name == "result":
            response, python_object = handler.read_from_serialized(
                lattice_data["results_filename"]
            )
            return LatticeFileResponse(data=str(response), python_object=python_object)

        if name == "inputs":
            response, python_object = handler.read_from_serialized(lattice_data["inputs_filename"])
            return LatticeFileResponse(data=response, python_object=python_object)

        elif name == "function_string":
            response = handler.read_from_text(lattice_data["function_string_filename"])
            return LatticeFileResponse(data=response)

        elif name == "executor":
            executor_name = lattice_data["executor"]
            executor_data = json.loads(lattice_data["executor_data"])

            return LatticeExecutorResponse(
                executor_name=executor_name, executor_details=executor_data
            )

        elif name == "workflow_executor":
            executor_name = lattice_data["workflow_executor"]
            executor_data = json.loads(lattice_data["workflow_executor_data"])

            return LatticeWorkflowExecutorResponse(
                workflow_executor_name=executor_name, workflow_executor_details=executor_data
            )

        elif name == "error":
            response = handler.read_from_text(lattice_data["error_filename"])
            return LatticeFileResponse(data=response)

        elif name == "function":
            response, python_object = handler.read_from_serialized(
                lattice_data["function_filename"]
            )
            return LatticeFileResponse(data=response, python_object=python_object)

        else:
            return LatticeFileResponse(data=None)


@routes.get("/{dispatch_id}/sublattices", response_model=SubLatticeDetailResponse)
def get_sub_lattice(
    sort_by: Optional[SubLatticeSortBy] = Query(default=SubLatticeSortBy.RUNTIME),
    sort_direction: Optional[SortDirection] = Query(default=SortDirection.DESCENDING),
    dispatch_id: uuid.UUID = Path(title="dispatch id"),
):
    """Get All Sub Lattices

    Args:
        req: Dispatch ID

    Returns:
        List of Sub Lattices details
    """
    with Session(db.engine) as session:
        lattice = Lattices(session)
        data = lattice.get_sub_lattice_details(
            dispatch_id=dispatch_id, sort_by=sort_by, sort_direction=sort_direction
        )
        return SubLatticeDetailResponse(sub_lattices=data)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

import re
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from covalent._shared_files.schemas.result import ResultSchema

range_regex = "bytes=([0-9]+)-([0-9]*)"
range_pattern = re.compile(range_regex)

digest_regex = "(sha|sha-256)=([0-9a-f]+)"
digest_pattern = re.compile(digest_regex)


class DispatchAssetKey(str, Enum):
    result = "result"
    error = "error"


class LatticeAssetKey(str, Enum):
    workflow_function = "workflow_function"
    workflow_function_string = "workflow_function_string"
    doc = "doc"
    inputs = "inputs"
    named_args = "named_args"
    named_kwargs = "named_kwargs"
    hooks = "hooks"
    cova_imports = "cova_imports"
    lattice_imports = "lattice_imports"


class ElectronAssetKey(str, Enum):
    function = "function"
    function_string = "function_string"
    output = "output"
    value = "value"
    hooks = "hooks"
    error = "error"
    stdout = "stdout"
    stderr = "stderr"
    qelectron_db = "qelectron_db"


class ExportResponseSchema(BaseModel):
    id: str
    status: str
    result_export: Optional[ResultSchema] = None


class AssetRepresentation(str, Enum):
    string = "string"
    b64pickle = "object"


class TargetDispatchStatus(str, Enum):
    running = "RUNNING"
    cancelled = "CANCELLED"


class DispatchStatusSetSchema(BaseModel):
    # The target status
    status: TargetDispatchStatus

    # For cancellation, an optional list of task ids to cancel
    task_ids: Optional[List] = []



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


class Order(str, enum.Enum):
    BEFORE = "before"
    AFTER = "after"


class FileSchemes(str, enum.Enum):
    File = "file"
    S3 = "s3"
    Blob = "https"
    GCloud = "gs"
    Globus = "globus"
    HTTP = "http"
    HTTPS = "https"
    FTP = "ftp"


class FileTransferStrategyTypes(str, enum.Enum):
    Shutil = "Shutil"
    Rsync = "rsync"
    HTTP = "http"
    S3 = "s3"
    GCloud = "gs"
    FTP = "ftp"
    GLOBUS = "globus"


SchemeToStrategyMap = {
    "file": FileTransferStrategyTypes.Shutil,
    "http": FileTransferStrategyTypes.HTTP,
    "https": FileTransferStrategyTypes.HTTP,
    "s3": FileTransferStrategyTypes.S3,
    "gs": FileTransferStrategyTypes.GCloud,
    "ftp": FileTransferStrategyTypes.FTP,
    "globus": FileTransferStrategyTypes.GLOBUS,
}


class FtCallDepReturnValue(str, enum.Enum):
    TO = "to"
    FROM = "from"
    FROM_TO = "from_to"



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
import logging
import os
import shutil
import subprocess
import sys
from configparser import ConfigParser
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from covalent._shared_files.config import set_config
from covalent.executor import _executor_manager

logger = logging.getLogger()
logger.setLevel(logging.ERROR)
handler = logging.StreamHandler(sys.stderr)
logger.addHandler(handler)
logger.propagate = False


def get_executor_module(executor_name: str) -> ModuleType:
    """
    Get the executor module from the executor name

    Args:
        executor_name: Name of the executor

    Returns:
        The executor module

    """
    return importlib.import_module(
        _executor_manager.executor_plugins_map[executor_name].__module__
    )


def validate_options(
    ExecutorPluginDefaults, ExecutorInfraDefaults, executor_options: Dict[str, str]
) -> None:
    """
    Validate the options passed to the CRM

    Args:
        ExecutorPluginDefaults: Executor plugin defaults validation class
        ExecutorInfraDefaults: Executor infra defaults validation class
        executor_options: Options passed to the CRM

    Raises:
        pydantic.ValidationError: If the options are invalid

    """
    # Validating the passed options:

    plugin_attrs = list(ExecutorPluginDefaults.schema()["properties"].keys())
    infra_attrs = list(ExecutorInfraDefaults.schema()["properties"].keys())

    plugin_params = {k: v for k, v in executor_options.items() if k in plugin_attrs}
    infra_params = {k: v for k, v in executor_options.items() if k in infra_attrs}

    # Validate options
    ExecutorPluginDefaults(**plugin_params)
    ExecutorInfraDefaults(**infra_params)


def get_plugin_settings(
    ExecutorPluginDefaults, ExecutorInfraDefaults, executor_options: Dict
) -> Dict:
    """Get plugin settings.

    Args:
        ExecutorPluginDefaults: Executor plugin defaults validation class.
        ExecutorInfraDefaults: Executor infra defaults validation class.
        executor_options: Resource provisioning options passed to the CRM.

    Returns:
        Dictionary of plugin settings.

    """
    plugin_settings = ExecutorPluginDefaults.schema()["properties"]
    infra_settings = ExecutorInfraDefaults.schema()["properties"]

    settings_dict = {
        key: {
            "required": "No",
            "default": value["default"],
            "value": value["default"],
        }
        if "default" in value
        else {"required": "Yes", "default": None, "value": None}
        for key, value in plugin_settings.items()
    }
    for key, value in infra_settings.items():
        if "default" in value:
            settings_dict[key] = {
                "required": "No",
                "default": value["default"],
                "value": value["default"],
            }
        else:
            settings_dict[key] = {"required": "Yes", "default": None, "value": None}

    if executor_options:
        for key, value in executor_options.items():
            try:
                settings_dict[key]["value"] = value
            except Exception:
                logger.error(f"No such option '{key}'. Use --help for available options")
                sys.exit(1)

    return settings_dict


class CloudResourceManager:
    """
    Base cloud resource manager class
    """

    def __init__(
        self,
        executor_name: str,
        executor_module_path: str,
        options: Optional[Dict[str, str]] = None,
    ):
        self.executor_name = executor_name
        self.executor_tf_path = str(
            Path(executor_module_path).expanduser().resolve() / "assets" / "infra"
        )

        # Includes both plugin and infra options
        self.executor_options = options

        # Importing validation classes from the executor module
        module = get_executor_module(executor_name)
        self.ExecutorPluginDefaults = getattr(module, "ExecutorPluginDefaults")
        self.ExecutorInfraDefaults = getattr(module, "ExecutorInfraDefaults")

        if self.executor_options:
            validate_options(
                self.ExecutorPluginDefaults, self.ExecutorInfraDefaults, self.executor_options
            )

        self.plugin_settings = get_plugin_settings(
            self.ExecutorPluginDefaults, self.ExecutorInfraDefaults, self.executor_options
        )

        self._terraform_log_env_vars = {
            "TF_LOG": "ERROR",
            "TF_LOG_PATH": os.path.join(self.executor_tf_path, "terraform-error.log"),
        }

    def _poll_process(self, process: subprocess.Popen, print_callback: Callable) -> int:
        """
        Print the stdout from the subprocess to console

        Args:
            process: Python subprocess whose stdout is to be printed to screen.
            print_callback: Callback function to print the stdout.

        Returns:
            Return code of the process.

        """
        while (returncode := process.poll()) is None:
            if print_callback:
                print_callback(process.stdout.readline())
        return returncode

    def _parse_terraform_error_log(self) -> List[str]:
        """Parse the terraform error logs.

        Returns:
            List of lines in the terraform error log.

        """
        with open(Path(self.executor_tf_path) / "terraform-error.log", "r", encoding="UTF-8") as f:
            lines = f.readlines()
        for _, line in enumerate(lines):
            error_index = line.strip().find("error:")
            if error_index != -1:
                error_message = line.strip()[error_index + len("error:") :]
                logger.error(error_message)
        return lines

    def _terraform_error_validator(self, tfstate_path: str) -> bool:
        """
        Terraform error validator checks whether any terraform-error.log files existence and validate last line.
        Args: None
        Return:
            up    - if terraform-error.log is empty and tfstate exists.
            *up   - if terraform-error.log is not empty and 'On deploy' at last line.
            down  - if terraform-error.log is empty and tfstate file not exists.
            *down - if terraform-error.log is not empty and 'On destroy' at last line.
        """
        tf_error_file = os.path.join(self.executor_tf_path, "terraform-error.log")
        if os.path.exists(tf_error_file) and os.path.getsize(tf_error_file) > 0:
            with open(tf_error_file, "r", encoding="UTF-8") as error_file:
                indicator = error_file.readlines()[-1]
                if indicator == "On deploy":
                    return "*up"
                elif indicator == "On destroy":
                    return "*down"
        return "up" if os.path.exists(tfstate_path) else "down"

    def _get_resource_status(
        self,
        proc: subprocess.Popen,
        cmd: str,
    ) -> str:
        """
        Get resource status will return current status of plugin based on terraform-error.log and tfstate file.
        Args:
            proc  : subprocess.Popen - To read stderr from Popen.communicate.
            cmd   : command for executing terraform scripts.
        Returns:
            status: str - status of plugin
        """

        cmds = cmd.split(" ")
        tfstate_path = cmds[-1].split("=")[-1]

        returncode = self._poll_process(proc, print_callback=None)
        stderr = proc.stderr.read()
        if returncode != 0 and "No state file was found!" not in stderr:
            print(
                "Unable to get resource status due to the following error:\n\n",
                stderr,
                file=sys.stderr,
            )

        return self._terraform_error_validator(tfstate_path=tfstate_path)

    def _log_error_msg(self, cmd) -> None:
        """
        Log error msg with valid command to terraform-erro.log
        Args: cmd: str - terraform-error.log file path.
        """
        with open(
            Path(self.executor_tf_path) / "terraform-error.log", "a", encoding="UTF-8"
        ) as file:
            if any(tf_cmd in cmd for tf_cmd in ["init", "plan", "apply"]):
                file.write("\nOn deploy")
            elif "destroy" in cmd:
                file.write("\nOn destroy")

    def _run_in_subprocess(
        self,
        cmd: str,
        env_vars: Optional[Dict[str, str]] = None,
        print_callback: Optional[Callable] = None,
    ) -> Union[None, str]:
        """
        Run the `cmd` in a subprocess shell with the env_vars set in the process's new environment

        Args:
            cmd: Command to execute in the subprocess
            env_vars: Dictionary of environment variables to set in the processes execution environment

        Returns:
            Union[None, str]
                - For 'covalent deploy status'
                    returns status of the deployment
                - Others
                    return None
        """
        if not shutil.which("git"):
            self._log_error_msg(cmd=cmd)
            logger.error("Git not found on the system.")
            sys.exit(1)

        env_vars = env_vars or {}
        env_vars.update({"PATH": os.environ["PATH"]})

        proc = subprocess.Popen(
            args=cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.executor_tf_path,
            universal_newlines=True,
            shell=True,
            env=env_vars,
        )

        if "state list -state" in cmd:
            return self._get_resource_status(proc=proc, cmd=cmd)

        returncode = self._poll_process(proc, print_callback)

        if returncode != 0:
            self._log_error_msg(cmd=cmd)

            _, stderr = proc.communicate()
            raise subprocess.CalledProcessError(
                returncode=returncode,
                cmd=cmd,
                stderr=self._parse_terraform_error_log(),
                output=stderr,
            )

        return None

    def _update_config(self, tf_executor_config_file: str) -> None:
        """
        Update covalent configuration with the executor
        config values as obtained from terraform

        Args:
            tf_executor_config_file: Path to the terraform executor config file

        Returns:
            None

        """
        # Puts the plugin options in covalent's config
        executor_config = ConfigParser()
        executor_config.read(tf_executor_config_file)

        filtered_executor_config = {
            k: v if v != "null" else None for k, v in executor_config[self.executor_name].items()
        }

        validated_config = self.ExecutorPluginDefaults(**filtered_executor_config).dict()

        for key, value in validated_config.items():
            set_config({f"executors.{self.executor_name}.{key}": value})
            self.plugin_settings[key]["value"] = value

    def _validation_docker(self) -> None:
        if not shutil.which("docker"):
            logger.error("Docker not found on system")
            sys.exit()

    def _get_tf_path(self) -> str:
        """
        Get the terraform path

        Args:
            None

        Returns:
            Path to terraform executable

        """
        if terraform := shutil.which("terraform"):
            result = subprocess.run(
                ["terraform --version"],
                shell=True,
                capture_output=True,
                text=True,
                check=True,
            )
            version = result.stdout.split("v", 1)[1][:3]
            if float(version) < 1.4:
                logger.error(
                    "Old version of terraform found. Please update it to version greater than 1.3"
                )
                sys.exit(1)
            return terraform

        logger.error("Terraform not found on system")
        sys.exit(1)

    def _get_tf_statefile_path(self) -> str:
        """
        Get the terraform state file path

        Args:
            None

        Returns:
            Path to terraform state file

        """
        # Saving in a directory which doesn't get deleted on purge
        return str(Path(self.executor_tf_path) / "terraform.tfstate")

    def up(self, print_callback: Callable, dry_run: bool = True) -> None:
        """
        Spin up executor resources with terraform

        Args:
            dry_run: If True, only run terraform plan and not apply

        Returns:
            None

        """
        terraform = self._get_tf_path()
        self._validation_docker()
        tfvars_file = Path(self.executor_tf_path) / "terraform.tfvars"
        tf_executor_config_file = Path(self.executor_tf_path) / f"{self.executor_name}.conf"

        tf_init = " ".join([terraform, "init"])
        tf_plan = " ".join([terraform, "plan", "-out", "tf.plan"])
        tf_apply = " ".join([terraform, "apply", "tf.plan"])
        terraform_log_file = self._terraform_log_env_vars["TF_LOG_PATH"]

        if Path(terraform_log_file).exists():
            Path(terraform_log_file).unlink()

        # Run `terraform init`
        self._run_in_subprocess(
            cmd=tf_init,
            env_vars=self._terraform_log_env_vars,
            print_callback=print_callback,
        )

        # Setup terraform infra variables as passed by the user
        tf_vars_env_dict = os.environ.copy()

        # Write the default values to the terraform.tfvars file
        infra_settings = self.ExecutorInfraDefaults.schema()["properties"]
        with open(tfvars_file, "w", encoding="utf-8") as f:
            for key, value in infra_settings.items():
                if "default" in value:
                    tf_vars_env_dict[f"TF_VAR_{key}"] = value["default"]

                    if value["default"]:
                        f.write(f'{key}={self._convert_to_tfvar(value["default"])}\n')

        # Overwrite the default values with the user passed values
        if self.executor_options:
            with open(tfvars_file, "w", encoding="utf-8") as f:
                for key, value in self.executor_options.items():
                    tf_vars_env_dict[f"TF_VAR_{key}"] = value

                    # Write whatever the user has passed to the terraform.tfvars file
                    f.write(f'{key}="{value}"\n')

        # Run `terraform plan`
        self._run_in_subprocess(
            cmd=tf_plan,
            env_vars=self._terraform_log_env_vars,
            print_callback=print_callback,
        )

        # Create infrastructure as per the plan
        # Run `terraform apply`
        if not dry_run:
            self._run_in_subprocess(
                cmd=tf_apply,
                env_vars=tf_vars_env_dict.update(self._terraform_log_env_vars),
                print_callback=print_callback,
            )

            # Update covalent executor config based on Terraform output
            self._update_config(tf_executor_config_file)

        if Path(terraform_log_file).exists() and os.path.getsize(terraform_log_file) == 0:
            Path(terraform_log_file).unlink()

    def down(self, print_callback: Callable) -> None:
        """
        Teardown previously spun up executor resources with terraform.

        Args:
            print_callback: Callback function to print output.

        Returns:
            None

        """
        terraform = self._get_tf_path()
        self._validation_docker()
        tf_state_file = self._get_tf_statefile_path()
        tfvars_file = Path(self.executor_tf_path) / "terraform.tfvars"
        terraform_log_file = self._terraform_log_env_vars["TF_LOG_PATH"]

        tf_destroy = " ".join(
            [
                "TF_CLI_ARGS=-no-color",
                "TF_LOG=ERROR",
                f"TF_LOG_PATH={terraform_log_file}",
                terraform,
                "destroy",
                "-auto-approve",
            ]
        )
        if Path(terraform_log_file).exists():
            Path(terraform_log_file).unlink()

        # Run `terraform destroy`
        self._run_in_subprocess(
            cmd=tf_destroy,
            print_callback=print_callback,
            env_vars=self._terraform_log_env_vars,
        )

        if Path(tfvars_file).exists():
            Path(tfvars_file).unlink()

        if Path(terraform_log_file).exists() and os.path.getsize(terraform_log_file) == 0:
            Path(terraform_log_file).unlink()

        if Path(tf_state_file).exists():
            Path(tf_state_file).unlink()
            if Path(f"{tf_state_file}.backup").exists():
                Path(f"{tf_state_file}.backup").unlink()

    def status(self) -> None:
        """
        Get the status of the spun up executor resources

        TODO: Return the list of resources being managed by terraform, i.e.
        if empty, then either the resources have not been created or
        have been destroyed already.

        Args:
            None

        Returns:
            None

        """
        terraform = self._get_tf_path()
        self._validation_docker()
        tf_state_file = self._get_tf_statefile_path()

        tf_state = " ".join([terraform, "state", "list", f"-state={tf_state_file}"])

        # Run `terraform state list`
        return self._run_in_subprocess(cmd=tf_state, env_vars=self._terraform_log_env_vars)

    @staticmethod
    def _convert_to_tfvar(value: Any) -> Any:
        """
        Convert the value to a string that can be parsed as a terraform variable.

        Args:
            value: Value to convert

        Returns:
            Converted value

        """
        if value is True:
            return "true"
        if value is False:
            return "false"
        if value is None:
            return "null"
        if isinstance(value, str):
            return f'"{value}"'
        if isinstance(value, Sequence):
            values = [CloudResourceManager._convert_to_tfvar(v) for v in value]
            return f"[{', '.join(values)}]"

        return str(value)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Handlers for the executor proxy """


from covalent._shared_files import logger
from covalent._shared_files.util_classes import Status
from covalent_dispatcher._core.data_modules import lattice as lattice_query_module

from .. import data_manager as datasvc
from ..data_modules import job_manager
from .utils import get_executor

app_log = logger.app_log
log_stack_info = logger.log_stack_info


async def get_cancel_requested(dispatch_id: str, task_id: int):
    """
    Query the database for the task's cancellation status

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice

    Return(s)
        Canellation status of the task
    """

    app_log.debug(f"Get _handle_requested for task {dispatch_id}:{task_id}")
    job_records = await job_manager.get_jobs_metadata(dispatch_id, [task_id])
    app_log.debug(f"Job record: {job_records[0]}")
    return job_records[0]["cancel_requested"]


async def get_version_info(dispatch_id: str, task_id: int):
    """
    Query the database for the dispatch version information

    Arg:
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice

    Returns:
        {"python": python_version, "covalent": covalent_version}
    """

    data = await lattice_query_module.get(dispatch_id, ["python_version", "covalent_version"])

    return {
        "python": data["python_version"],
        "covalent": data["covalent_version"],
    }


async def get_job_status(dispatch_id: str, task_id: int) -> Status:
    """
    Queries the job state for (dispatch_id, task_id)

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice

    Return(s)
        Status
    """
    app_log.debug(f"Get for task {dispatch_id}:{task_id}")
    job_records = await job_manager.get_jobs_metadata(dispatch_id, [task_id])
    app_log.debug(f"Job record: {job_records[0]}")
    return Status(job_records[0]["status"])


async def put_job_handle(dispatch_id: str, task_id: int, job_handle: str) -> bool:
    """
    Store the job handle of the task returned by the backend in the database

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice
        job_handle: Unique identifier of the task returned by the execution backend

    Return(s)
        True
    """
    app_log.debug(f"Put job_handle for executor {dispatch_id}:{task_id}")
    await job_manager.set_job_handle(dispatch_id, task_id, job_handle)
    return True


async def put_job_status(dispatch_id: str, task_id: int, status: Status) -> bool:
    """
    Mark the job for (dispatch_id, task_id) as cancelled

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice
        job_status: A `Status` type representing the job status

    Return(s)
        True
    """
    app_log.debug(f"Put cancel result for task {dispatch_id}:{task_id}")
    executor_attrs = await datasvc.electron.get(
        dispatch_id, task_id, ["executor", "executor_data"]
    )
    selected_executor = [executor_attrs["executor"], executor_attrs["executor_data"]]
    executor = get_executor(task_id, selected_executor, None, None)
    if executor.validate_status(status):
        await job_manager.set_job_status(dispatch_id, task_id, str(status))
        return True
    else:
        return False



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing post-processing related functions."""

from builtins import list
from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Dict, List, Set, Union

from .._shared_files import logger
from .._shared_files.context_managers import active_lattice_manager
from .._shared_files.defaults import (
    DefaultMetadataValues,
    postprocess_prefix,
    prefix_separator,
    sublattice_prefix,
)
from .transport import _TransportGraph

if TYPE_CHECKING:
    from .electron import Electron

DEFAULT_METADATA_VALUES = asdict(DefaultMetadataValues())

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class Postprocessor:
    def __init__(self, lattice) -> None:
        self.lattice = lattice

    def _is_postprocessable_node(self, tg: _TransportGraph, node_id: int) -> bool:
        """Filter nodes that should be included in postprocessing.

        Args:
            tg: Transport graph of the lattice.
            node_id: Node id of the node to be checked.

        Returns:
            True if the node should be included in postprocessing, False otherwise.

        """
        node_name = tg.get_node_value(node_id, "name")
        return bool(
            not node_name.startswith(prefix_separator) or node_name.startswith(sublattice_prefix)
        )

    def _filter_electrons(self, tg: _TransportGraph, bound_electrons: List) -> List:
        """Filter bound electrons for ones that should not be included in postprocessing.

        Args:
            tg: Transport graph of the lattice.
            bound_electrons: List of bound electrons.

        Returns:
            List of bound electrons that should be included in postprocessing.

        """
        filtered_node_ids = [
            node_id for node_id in tg._graph.nodes if self._is_postprocessable_node(tg, node_id)
        ]
        return [bound_electrons[node_id] for node_id in filtered_node_ids]

    def _postprocess(self, *ordered_node_outputs: List[Any]) -> Any:
        """
        Post processing function to be called after the lattice execution.
        This takes care of executing statements that were not an electron
        but were inside the lattice's function. It also replaces any calls
        to an electron with the result of that electron execution, hence
        preventing a local execution of electron's function.
        Note: Here `node_outputs` is used instead of `electron_outputs`
        since an electron can be called multiple times with possibly different
        arguments, but every time it's called, it will be executed as a separate node.
        Thus, output of every node is used.

        Args:
            ordered_node_outputs: List of outputs of every node in the lattice.

        Returns:
            result: The result of the lattice function.

        """
        with active_lattice_manager.claim(self.lattice):
            self.lattice.post_processing = True
            self.lattice.electron_outputs = list(ordered_node_outputs)
            inputs = self.lattice.inputs.get_deserialized()
            args = inputs["args"]
            kwargs = inputs["kwargs"]
            workflow_function = self.lattice.workflow_function.get_deserialized()
            result = workflow_function(*args, **kwargs)
            self.lattice.post_processing = False
            return result

    def _get_node_ids_from_retval(self, retval: Union["Electron", List, Dict]) -> Set[int]:
        """Get the list of electron node ids from the return value of a lattice function.

        Args:
            retval: Return value of a lattice function.

        Returns:
            List of electron node ids.

        """
        from .electron import Electron

        node_ids = []
        if isinstance(retval, Electron):
            node_ids.append(retval.node_id)
            app_log.debug(f"Preprocess: Encountered node {retval.node_id}")
        elif isinstance(retval, (list, tuple, set)):
            app_log.debug("Recursively preprocessing iterable")
            for e in retval:
                node_ids.extend(self._get_node_ids_from_retval(e))
        elif isinstance(retval, dict):
            app_log.debug("Recursively preprocessing dictionary")
            for _, v in retval.items():
                node_ids.extend(self._get_node_ids_from_retval(v))
        else:
            app_log.debug(f"Encountered primitive or unsupported type: {retval}")
            return []

        return set(node_ids)

    def _get_electron_metadata(self) -> Dict:
        """Get the metadata for the postprocess electron.

        Returns:
            Dictionary of metadata for the postprocess electron.

        """

        # This is already initialized and JSON-encoded at the
        # beginning of Lattice.build_graph
        pp_metadata = self.lattice.metadata.copy()

        pp_metadata["executor"] = pp_metadata.pop("workflow_executor")
        pp_metadata["executor_data"] = pp_metadata.pop("workflow_executor_data")

        return pp_metadata

    def add_exhaustive_postprocess_node(self, bound_electrons: Dict) -> None:
        """This function adds a postprocess node to the transport graph based on the 'exhaustive' algorithm.

        Args:
            bound_electrons: Dictionary of bound electrons with node_id as key and the electron object as the value. A bound is an electron that has been assigned a node_id and is part of a transport graph.

        Returns:
            None

        """
        from .electron import Electron

        with active_lattice_manager.claim(self.lattice):
            tg = self.lattice.transport_graph
            filtered_ordered_electrons = self._filter_electrons(tg, bound_electrons)
            pp_electron = Electron(
                function=self._postprocess, metadata=self._get_electron_metadata()
            )
            bound_pp = pp_electron(*filtered_ordered_electrons)
            tg.set_node_value(bound_pp.node_id, "name", postprocess_prefix)

    def _postprocess_recursively(self, retval, **referenced_outputs):
        from .electron import Electron

        if isinstance(retval, Electron):
            key = f"node:{retval.node_id}"
            return referenced_outputs[key]
        elif isinstance(retval, list):
            return list(
                map(lambda x: self._postprocess_recursively(x, **referenced_outputs), retval)
            )
        elif isinstance(retval, tuple):
            return tuple(
                map(lambda x: self._postprocess_recursively(x, **referenced_outputs), retval)
            )
        elif isinstance(retval, set):
            return {self._postprocess_recursively(x, **referenced_outputs) for x in retval}
        elif isinstance(retval, dict):
            return {
                k: self._postprocess_recursively(v, **referenced_outputs)
                for k, v in retval.items()
            }
        else:
            return retval

    def add_reconstruct_postprocess_node(
        self, retval: Union["Electron", List, Dict], bound_electrons: Dict
    ):
        """This function adds a postprocess node to the transport graph based on the 'eager reconstruction' algorithm.

        Args:
            retval: Return value of the lattice function.
            bound_electrons: Dictionary of bound electrons with node_id as key and the electron object as the value. A bound is an electron that has been assigned a node_id and is part of a transport graph.

        Returns:
            None

        """
        from .electron import Electron, wait

        node_id_refs = self._get_node_ids_from_retval(retval)
        referenced_electrons = {}
        for node_id in node_id_refs:
            key = f"node:{node_id}"
            referenced_electrons[key] = bound_electrons[node_id]

        with active_lattice_manager.claim(self.lattice):
            pp_metadata = self._get_electron_metadata()
            pp_electron = Electron(function=self._postprocess_recursively, metadata=pp_metadata)

            # Add pp_electron to the graph -- this will also add a
            # parameter node in case retval is not a single electron
            bound_pp = pp_electron(retval, **referenced_electrons)

            # Edit pp electron name
            self.lattice.transport_graph.set_node_value(
                bound_pp.node_id, "name", f"{postprocess_prefix}reconstruct"
            )

            # Wait for non-referenced electrons
            wait_parents = [v for k, v in bound_electrons.items() if k not in node_id_refs]
            wait(child=bound_pp, parents=wait_parents)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod, abstractproperty


class BaseQClient(ABC):
    @abstractmethod
    def submit(self, qscripts, executors, qelectron_info, qnode_specs):
        raise NotImplementedError

    @abstractmethod
    def get_results(self, batch_id):
        raise NotImplementedError

    @abstractproperty
    def selector(self):
        raise NotImplementedError

    @abstractproperty
    def database(self):
        raise NotImplementedError

    # The following methods are abstract because the qserver
    # is expecting serialized inputs and will be sending
    # back serialized outputs, thus even if these methods
    # essentially just pass through, for e.g in the LocalQClient's
    # case, they are still to be implemented by the child class and
    # should use the same seriliazing/deserializing method as is being
    # used by the equivalent qserver.
    @abstractmethod
    def serialize(self, obj):
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, ser_obj):
        raise NotImplementedError



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quantum Server Implementation: Handles the async execution of quantum circuits.
"""

import datetime
import uuid
from asyncio import Task
from typing import Callable, List, Tuple

from pennylane.tape import QuantumScript

from ..._shared_files.qelectron_utils import (
    cloudpickle_deserialize,
    cloudpickle_serialize,
    select_first_executor,
)
from ..._shared_files.qinfo import QElectronInfo, QNodeSpecs
from ...executor.utils import get_context
from ..qcluster.base import AsyncBaseQCluster, BaseQExecutor
from .database import Database
from .utils import CircuitInfo, get_cached_executor, get_circuit_id


class FuturesTable:
    """
    Container for async task futures corresponding to a sub-batch of executing
    qscripts, as identified by a batch UUID.
    """

    def __init__(self):
        self._ef_pairs = {}

    def add_executor_future_pairs(
        self,
        executor_future_pairs: List[Tuple[BaseQExecutor, Task]],
        submission_order: List[int],
    ) -> str:
        """
        Add a list of futures to the table and return a corresponding UUID.
        """
        batch_id = str(uuid.uuid4())
        self._ef_pairs[batch_id] = (executor_future_pairs, submission_order)
        return batch_id

    def pop_executor_future_pairs(
        self,
        batch_id: str,
    ) -> Tuple[List[Tuple[BaseQExecutor, Task]], List[int]]:
        """
        Retrieve a list of futures from the table using a UUID.
        """
        return self._ef_pairs.pop(batch_id)


class QServer:
    """
    Initialize a QServer instance with a given selector function.
    """

    # def __init__(self, selector: BaseQSelector = None) -> None:
    def __init__(self, selector: Callable = None) -> None:
        self.futures_table = FuturesTable()

        # self._selector = selector or SimpleSelector(selector_function=select_first_executor)
        self._selector = selector or select_first_executor
        self._database = Database()

    @property
    def selector(self):
        """
        Executor selector function for the Quantum server.
        """
        return self._selector

    @selector.setter
    def selector(self, ser_selector):
        self._selector = self.deserialize(ser_selector)

    @property
    def database(self):
        """Return the database for reading."""
        return self.serialize(self._database)

    def select_executors(
        self,
        qscripts: List[QuantumScript],
        executors: List[BaseQExecutor],
        qnode_specs: QNodeSpecs,
    ):
        """
        Links qscripts with an executor
        based on the self.selector function
        """

        linked_executors = []
        for qscript in qscripts:
            selected_executor = self.selector(qscript, executors)

            # Use cached executor.
            selected_executor = get_cached_executor(**selected_executor.dict())

            if isinstance(selected_executor, AsyncBaseQCluster):
                # Apply QCluster's selector as well.
                qcluster = selected_executor
                selected_executor = qcluster.get_selector()(qscript, qcluster.executors)

                # Use cached executor.
                selected_executor = get_cached_executor(**selected_executor.dict())

            # This is the only place where the qnode_specs are set.
            selected_executor.qnode_specs = qnode_specs.copy()

            # An example `linked_executors` will look like:
            # [exec_4, exec_4, exec_2, exec_3]
            # Their indices corresponding to the indices of `qscripts`.
            linked_executors.append(selected_executor)

        return linked_executors

    def submit_to_executors(
        self,
        qscripts: List[QuantumScript],
        linked_executors: List[BaseQExecutor],
        qelectron_info: QElectronInfo,
    ):
        """
        Generates futures for scheduled execution
        of qscripts on respective executors
        """

        # Since we will be modifying the qscripts list (or sometimes tuple).
        qscripts = list(qscripts).copy()

        submission_order = []
        executor_qscript_sub_batch_pairs = []
        for i, qscript in enumerate(qscripts):
            if qscript is None:
                continue

            # Generate a sub batch of qscripts to be executed on the same executor
            qscript_sub_batch = [linked_executors[i], {i: qscript}]

            # The qscript submission order is stored in this list to ensure that
            # the final result is recombined correctly, even if task-circuit
            # correspondence is not one-to-one. See, for example, PR #13.
            submission_order.append(i)

            for j in range(i + 1, len(qscripts)):
                if linked_executors[i] == linked_executors[j]:
                    qscript_sub_batch[1][j] = qscripts[j]
                    qscripts[j] = None
                    submission_order.append(j)

            # An example `qscript_sub_batch` will look like:
            # [exec_4, {0: qscript_1, 3: qscript_4}]

            executor_qscript_sub_batch_pairs.append(qscript_sub_batch)

        # An example `executor_qscript_sub_batch_pairs` will look like:
        # [
        #    [exec_4, {0: qscript_1, 3: qscript_4}],
        #    [exec_2, {2: qscript_3}],
        #    [exec_3, {4: qscript_5}],
        # ]

        # Generating futures from each executor:
        executor_future_pairs = []
        for executor, qscript_sub_batch in executor_qscript_sub_batch_pairs:
            executor.qelectron_info = qelectron_info.copy()
            qscript_futures = executor.batch_submit(qscript_sub_batch.values())

            futures_dict = dict(zip(qscript_sub_batch.keys(), qscript_futures))
            # An example `futures_dict` will look like:
            # {0: future_1, 3: future_4}

            executor_future_pairs.append([executor, futures_dict])

        # An example `executor_future_pairs` will look like:
        # [[exec_4, {0: future_1, 3: future_4}], [exec_2, {2: future_3}], [exec_3, {4: future_5}]]

        return executor_future_pairs, submission_order

    def submit(
        self,
        qscripts: List[QuantumScript],
        executors: List[BaseQExecutor],
        qelectron_info: QElectronInfo,
        qnode_specs: QNodeSpecs,
    ):
        # pylint: disable=too-many-locals
        """
        Submit a list of QuantumScripts to the server for execution.

        Args:
            qscripts: A list of QuantumScripts to run.
            executors: The executors to choose from to use for running the QuantumScripts.
            qelectron_info: Information about the qelectron as provided by the user.
            qnode_specs: Specifications of the qnode.

        Returns:
            str: A UUID corresponding to the batch of submitted QuantumScripts.
        """

        # Get current electron's context
        context = get_context()

        # Get qelectron info, qnode specs, quantum scripts, and executors
        qelectron_info = self.deserialize(qelectron_info)
        qnode_specs = self.deserialize(qnode_specs)
        qscripts = self.deserialize(qscripts)
        executors = self.deserialize(executors)

        # Generate a list of executors for each qscript.
        linked_executors = self.select_executors(qscripts, executors, qnode_specs)

        # Assign qscript sub-batches to unique executors.
        executor_future_pairs, submission_order = self.submit_to_executors(
            qscripts, linked_executors, qelectron_info
        )

        # Get batch ID for N qscripts being async-executed on M <= N executors.
        batch_id = self.futures_table.add_executor_future_pairs(
            executor_future_pairs, submission_order
        )

        # Storing the qscripts, executors, and metadata in the database
        batch_time = str(datetime.datetime.now())
        key_value_pairs = [[], []]

        for i, qscript in enumerate(qscripts):
            circuit_id = get_circuit_id(batch_id, i)
            key_value_pairs[0].append(circuit_id)
            circuit_info = CircuitInfo(
                electron_node_id=context.task_id,
                dispatch_id=context.dispatch_id,
                circuit_name=qelectron_info.name,
                circuit_description=qelectron_info.description,
                circuit_diagram=qscript.draw(),
                qnode_specs=qnode_specs,
                qexecutor=linked_executors[i],
                save_time=batch_time,
                circuit_id=circuit_id,
                qscript=qscript.graph.serialize() if linked_executors[i].persist_data else None,
            ).dict()

            key_value_pairs[1].append(circuit_info)

        # An example `key_value_pairs` will look like:
        # [
        #     [
        #         "circuit_0-uuid",
        #         "circuit_1-uuid",
        #
        #         ...
        #     ],
        #     [
        #         {"electron_node_id": "node_1",
        #          "dispatch_id": "uuid",
        #          "circuit_name": "qscript_name",
        #          "circuit_description": "qscript_description",
        #          "qnode_specs": {"qnode_specs": "specs"},
        #          "qexecutor": "executor_1",
        #          "save_time": "2021-01-01 00:00:00",
        #          "circuit_id": "circuit_0-uuid",
        #          "qscript": "qscript_1"},

        #         {"electron_node_id": "node_1",
        #          "dispatch_id": "uuid",
        #          "circuit_name": "qscript_name",
        #          "circuit_description": "qscript_description",
        #          "qnode_specs": {"qnode_specs": "specs"},
        #          "qexecutor": "executor_2",
        #          "save_time": "2021-01-01 00:00:00",
        #          "circuit_id": "circuit_1-uuid",
        #          "qscript": "qscript_2"},
        #
        #         ...
        #     ],
        # ]

        self._database.set(
            *key_value_pairs, dispatch_id=context.dispatch_id, node_id=context.task_id
        )

        return batch_id

    def get_results(self, batch_id):
        # pylint: disable=too-many-locals
        """
        Retrieve the results of previously submitted QuantumScripts from the server.

        Args:
            batch_id: The UUID corresponding to the batch of submitted QuantumScripts.

        Returns:
            List: An ordered list of results for the submitted QuantumScripts.
        """

        # Get current electron's context
        context = get_context()

        results_dict = {}
        key_value_pairs = [[], []]
        executor_future_pairs, submission_order = self.futures_table.pop_executor_future_pairs(
            batch_id
        )

        # ids of (e)xecutor_(f)uture_(p)airs, hence `idx_efp`
        qscript_submission_index = 0
        for idx_efp, (executor, futures_sub_batch) in enumerate(executor_future_pairs):
            result_objs = executor.batch_get_results(futures_sub_batch.values())

            # Adding results according to the order of the qscripts
            # ids of (f)utures_(s)ub_(b)atch, hence `idx_fsb`
            for idx_fsb, circuit_number in enumerate(futures_sub_batch.keys()):
                result_obj = result_objs[idx_fsb]

                # Expand `result_obj` in case contains multiple circuits.
                # Loop through sub-results to store separately in db.
                for result_number, sub_result_obj in enumerate(result_obj.expand()):
                    qscript_number = submission_order[qscript_submission_index]

                    # Use tuple of integers for key to enable later multi-factor sort.
                    key = (qscript_number, circuit_number, result_number)
                    results_dict[key] = sub_result_obj.results[0]
                    qscript_submission_index += 1

                    # To store the results in the database
                    circuit_id = get_circuit_id(batch_id, circuit_number + result_number)
                    key_value_pairs[0].append(circuit_id)
                    key_value_pairs[1].append(
                        {
                            "execution_time": sub_result_obj.execution_time,
                            "result": sub_result_obj.results if executor.persist_data else None,
                            "result_metadata": sub_result_obj.metadata
                            if executor.persist_data
                            else None,
                        }
                    )

            # An example `key_value_pairs` will look like:
            # [
            #     [
            #         "result_circuit_0-uuid",
            #         "result_circuit_1-uuid",
            #     ],
            #     [
            #         {"execution_time": "2021-01-01 00:00:00",
            #          "result": [result_11, ...],
            #          "result_metadata": [{}, ...]},
            #
            #         {"execution_time": "2021-01-01 00:00:00",
            #          "result": [result_21, ...],
            #          "result_metadata": [{}, ...]},
            #     ],
            # ]

            # Deleting the futures once their results have been retrieved
            del executor_future_pairs[idx_efp][1]

            # After deletion of one `future_sub_batch`, the `executor_future_pairs` will look like:
            # [[exec_4], [exec_2, {2: future_3}], [exec_3, {4: future_5}]]

        self._database.set(
            *key_value_pairs, dispatch_id=context.dispatch_id, node_id=context.task_id
        )

        # An example `results_dict` will look like:
        # {0: result_1, 3: result_4, 2: result_3, 4: result_5}

        # Perform multi-factor sort on `results_dict`.
        batch_results = list(dict(sorted(results_dict.items())).values())

        # An example `batch_results` will look like:
        # [result_1, result_2, result_3, result_4, result_5]

        return self.serialize(batch_results)

    def serialize(self, obj):
        """
        Serialize an object.
        """
        return cloudpickle_serialize(obj)

    def deserialize(self, obj):
        """
        Deserialize an object.
        """
        return cloudpickle_deserialize(obj)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Any, Callable, List, Optional

import pennylane as qml
from pennylane import transforms
from pennylane.measurements import ExpectationMP
from pennylane.tape import QuantumTape

from .._workflow.qdevice import QEDevice
from .qelectron_utils import get_original_shots


def re_execute(
    results: Any,
    qnode: qml.QNode,
    tape: QuantumTape,
) -> Callable:
    """
    Send the QElectron QNode's result through a "shell" QNode to achieve correct typing
    and shape transformations.

    This is necessary because the QEDevice can not support 'passthru_devices' in
    its capabilities, since that would circumvent its custom `batch_execute` method.
    """

    def re_execute_wrapper(interface, *args, **kwargs):
        circuit = shell_circuit_factory(results, qnode, tape, interface)
        return circuit(*args, **kwargs)

    return re_execute_wrapper


def shell_circuit_factory(
    results: Any,
    qnode: qml.QNode,
    tape: QuantumTape,
    interface: Optional[str],
) -> qml.QNode:
    """
    Returns a circuit that does nothing but transform and return the results.
    """

    dev = shell_device_factory(results, qnode, tape, interface)

    # Define a dummy circuit that returns the original QNode's return value.
    @qml.qnode(dev, interface=qnode.interface, diff_method=qnode.diff_method)
    def _circuit(*_, **__):
        return tape._qfunc_output  # pylint: disable=protected-access

    return _circuit


def shell_device_factory(
    results: Any,
    qnode: qml.QNode,
    original_tape: QuantumTape,
    interface: Optional[str],
) -> qml.Device:
    """
    Returns an instance of a new class that inherits from the original QNode's device class.
    """

    default_capabilities = QEDevice.capabilities().copy()

    # Conditional override of the device's capabilities to accommodate interface.
    overriden_capabilities = _override_capabilities(results, interface, default_capabilities)

    class _ShellDevice(qnode.device.__class__):
        # pylint: disable=too-few-public-methods

        def batch_execute(self, circuits):
            """
            Override to return expected result.
            """
            return _reshape_for_interface(interface, circuits, results)

        def batch_transform(self, tape):
            """
            Ignore blank circuit and run batch transform on original tape.
            """

            # Conditional lifted from `qml.transforms.hamiltonian_expand`.
            if not (
                len(original_tape.measurements) != 1
                or not isinstance(original_tape.measurements[0].obs, qml.Hamiltonian)
                or not isinstance(original_tape.measurements[0], ExpectationMP)
            ):
                # Apply batch transform for Hamiltonian expvals.
                return transforms.hamiltonian_expand(original_tape)

            if original_tape.output_dim > 1:
                res = results if hasattr(results, "__len__") else [[r] for r in results]
                return [tape], lambda _, **kw: res

            # Apply identity transform.
            return [tape], lambda _, **kw: results

        @classmethod
        def capabilities(cls):
            """
            Copy the QElectron Device capabilities to avoid passthru devices and
            overload using `diff_method="backprop` and associated transform with
            default settings.
            """
            if overriden_capabilities is None:
                return QEDevice.capabilities().copy()
            return overriden_capabilities.copy()

    # Recreate device with original number of wires.
    dev = _ShellDevice(wires=qnode.device.num_wires, shots=1)

    # Use the `shots` property setter on `qml.Device` to set the shots or shot vector.
    dev.shots = get_original_shots(qnode.device)  # pylint: disable=attribute-defined-outside-init

    return dev


def _override_capabilities(
    results: Any, interface: str, default_capabilities: dict
) -> Optional[dict]:
    """
    Implements interface-based conditional overrides for device capabilities.
    """
    if interface in {None, "auto", "numpy", "autograd"}:
        # No override necessary.
        return default_capabilities

    # Require passthrough for some interfaces.
    if interface == "torch":
        if isinstance(results, list) or results.ndim == 0:
            return {
                "model": "qubit",
                "passthru_interface": "torch",
            }
        return default_capabilities

    if interface == "jax":
        return {
            "model": "qubit",
            "passthru_interface": "jax",
        }

    warnings.warn(f"Skipped capabilities override. No logic defined for '{interface}' interface")
    return default_capabilities


def _reshape_for_interface(interface: str, circuits: List[QuantumTape], results: Any):
    """
    Reshape or re-package the return value in an interface-specific way to satisfy
    expected shape or type requirements in Pennylane's execution pipeline.
    """

    if interface in {"auto", "numpy"}:
        # More than one circuit and more than one result.
        if len(circuits) > 1 and hasattr(results, "__len__") and len(results) > 1:
            return [[r] for r in results]

        # Results is a 0-dimensional array.
        if hasattr(results, "ndim") and results.ndim == 0:
            return [results]

        # No reshaping required.
        return results

    if interface in {None, "autograd", "torch", "jax"}:
        # Ensure result is a list.
        if isinstance(results, list):
            return results
        return [results]

    warnings.warn(f"Skipped reshaping result. No logic defined for '{interface}' interface")
    return results



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Covalent deploy CLI group."""

import subprocess
import sys
from functools import partial
from importlib import import_module
from pathlib import Path
from typing import Callable, Dict, Tuple

import click
from rich.console import Console
from rich.table import Table

from covalent.cloud_resource_manager.core import CloudResourceManager
from covalent.executor import _executor_manager

from .deploy_group_print_callbacks import ScrollBufferCallback

_TEMPLATE = "[bold green]{message} [default]\n {text}"


def get_crm_object(executor_name: str, options: Dict = None) -> CloudResourceManager:
    """
    Get the CloudResourceManager object.

    Returns:
        CloudResourceManager object.

    """
    executor_plugin = _executor_manager.executor_plugins_map[executor_name]
    executor_module_path = Path(import_module(executor_plugin.__module__).__file__).parent
    return CloudResourceManager(executor_name, executor_module_path, options)


def get_settings_table(crm: CloudResourceManager) -> Table:
    """Get resource provisioning settings table.

    Args:
        crm: CloudResourceManager object.

    Returns:
        Table with resource provisioning settings.

    """
    table = Table(title="Settings")
    table.add_column("Argument", justify="left")
    table.add_column("Value", justify="left")
    for argument in crm.plugin_settings:
        table.add_row(argument, str(crm.plugin_settings[argument]["value"]))
    return table


def get_up_help_table(crm: CloudResourceManager) -> Table:
    """Get resource provisioning help table.

    Args:
        crm: CloudResourceManager object.

    Returns:
        Table with resource provisioning help.

    """
    table = Table()
    table.add_column("Argument", justify="center")
    table.add_column("Required", justify="center")
    table.add_column("Default", justify="center")
    table.add_column("Current value", justify="center")
    for argument in crm.plugin_settings:
        table.add_row(
            argument,
            crm.plugin_settings[argument]["required"],
            str(crm.plugin_settings[argument]["default"]),
            str(crm.plugin_settings[argument]["value"]),
        )
    return table


def _run_command_and_show_output(
    _command: Callable[[Callable], None], _status_message: str, *, verbose: bool
) -> None:
    """Run the command and show the output in the console.

    This function handles execution and outputs from the `up` and `down` commands.

    Args:
        _command: command to run, e.g. `partial(crm.up, dry_run=dry_run)`
        _status_message: message to show in the console status bar, e.g. "Provisioning resources..."
        verbose: whether to show the full Terraform output or not.
    """
    console = Console(record=True)
    msg_template = _TEMPLATE.format(message=_status_message, text="{text}")

    with console.status(msg_template.format(text="")) as console_status:
        print_callback = ScrollBufferCallback(
            console=console,
            console_status=console_status,
            msg_template=msg_template,
            verbose=verbose,
        )

        try:
            _command(print_callback=print_callback)

        except subprocess.CalledProcessError as e:
            console_status.stop()
            click.echo(e.stdout)  # display error
            sys.exit(1)

        if not verbose:
            console_status.stop()
            if (complete_msg := print_callback.complete_msg) is not None:
                console.print("\n", complete_msg, style="bold green")


@click.group(invoke_without_command=True)
def deploy():
    """
    Covalent deploy group with options to:

    1. Spin resources up via `covalent deploy up <executor_name> <options>`.

    2. Tear resources down via `covalent deploy down <executor_name> <options>`.

    3. Show status of resources via `covalent deploy status <executor_name>`.

    4. Show status of all resources via `covalent deploy status`.

    """


@deploy.command(context_settings={"ignore_unknown_options": True})
@click.argument("executor_name", nargs=1)
@click.argument("vars", nargs=-1)
@click.option(
    "--help", "-h", is_flag=True, help="Get info on default and current values for resources."
)
@click.option("--dry-run", "-dr", is_flag=True, help="Get info on current parameter settings.")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show the full Terraform output when provisioning resources.",
)
def up(executor_name: str, vars: Dict, help: bool, dry_run: bool, verbose: bool) -> None:
    """Spin up resources corresponding to executor.

    Args:
        executor_name: Short name of executor to spin up.
        options: Options to pass to the Cloud Resource Manager when provisioning the resources.

    Returns:
        None

    Examples:
        $ covalent deploy up awsbatch --region=us-east-1 --instance-type=t2.micro
        $ covalent deploy up ecs
        $ covalent deploy up ecs --help
        $ covalent deploy up awslambda --verbose --region=us-east-1 --instance-type=t2.micro

    """

    cmd_options = {key[2:]: value for key, value in (var.split("=") for var in vars)}

    try:
        crm = get_crm_object(executor_name, cmd_options)
    except (KeyError, AttributeError):
        click.echo(
            click.style(
                f"Warning: '{executor_name}' is not a valid executor for deployment.",
                fg="yellow",
            )
        )
        sys.exit(1)

    if help:
        click.echo(Console().print(get_up_help_table(crm)))
        sys.exit(0)

    _command = partial(crm.up, dry_run=dry_run)
    _run_command_and_show_output(_command, "Provisioning resources...", verbose=verbose)


@deploy.command()
@click.argument("executor_name", nargs=1)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show the full Terraform output when spinning down resources.",
)
def down(executor_name: str, verbose: bool) -> None:
    """Teardown resources corresponding to executor.

    Args:
        executor_name: Short name of executor to spin up.

    Returns:
        None

    Examples:
        $ covalent deploy down awsbatch
        $ covalent deploy down ecs --verbose

    """

    try:
        crm = get_crm_object(executor_name)
    except (KeyError, AttributeError):
        click.echo(
            click.style(
                f"Warning: '{executor_name}' is not a valid executor for deployment.",
                fg="yellow",
            )
        )
        sys.exit(1)

    _command = partial(crm.down)
    _run_command_and_show_output(_command, "Destroying resources...", verbose=verbose)


# TODO - Color code status.
@deploy.command()
@click.argument("executor_names", nargs=-1, required=False)
def status(executor_names: Tuple[str]) -> None:
    """Show executor resource provision status.

    Args:
        executor_names: Short name(s) of executor to show status for.

    Returns:
        None

    Examples:
        $ covalent deploy status awsbatch
        $ covalent deploy status awsbatch ecs
        $ covalent deploy status

    """
    description = {
        "up": "Provisioned Resources.",
        "down": "No infrastructure provisioned.",
        "*up": "Warning: Provisioning error, retry 'up'.",
        "*down": "Warning: Teardown error, retry 'down'.",
    }
    if not executor_names:
        executor_names = [
            name
            for name in _executor_manager.executor_plugins_map
            if name not in ["dask", "local", "remote_executor"]
        ]
        click.echo(f"Installed executors: {', '.join(executor_names)}")

    table = Table()
    table.add_column("Executor", justify="center")
    table.add_column("Status", justify="center")
    table.add_column("Description", justify="center")

    invalid_executor_names = []
    for executor_name in executor_names:
        try:
            crm = get_crm_object(executor_name)
            crm_status = crm.status()
            table.add_row(executor_name, crm_status, description[crm_status])
        except (KeyError, AttributeError):
            # Added the AttributeError here as well in case the executor does not
            # have the ExecutorPluginDefaults or ExecutorInfraDefaults classes.
            invalid_executor_names.append(executor_name)

    click.echo(Console().print(table))

    if invalid_executor_names:
        click.echo(
            click.style(
                f"Warning: Invalid executors for deployment -> '{', '.join(invalid_executor_names)}'",
                fg="yellow",
            )
        )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class that defines the base executor template.
"""

import asyncio
import copy
import io
import json
import os
import queue
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Callable, ContextManager, Dict, Iterable, List, Optional, Union

import aiofiles

from covalent._shared_files.exceptions import TaskCancelledError
from covalent.executor.schemas import ResourceMap, TaskSpec
from covalent.executor.utils import Signals

from .._shared_files import TaskRuntimeError, logger
from .._shared_files.context_managers import active_dispatch_info_manager
from .._shared_files.util_classes import RESULT_STATUS, DispatchInfo, Status
from .schemas import TaskUpdate

app_log = logger.app_log
log_stack_info = logger.log_stack_info
TypeJSON = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class _AbstractBaseExecutor(ABC):
    """
    Private parent class for BaseExecutor and AsyncBaseExecutor

    Attributes:
        log_stdout: The path to the file to be used for redirecting stdout.
        log_stderr: The path to the file to be used for redirecting stderr.
        cache_dir: The location used for cached files in the executor.
        time_limit: time limit for the task
        retries: Number of times to retry execution upon failure

    """

    SUPPORTS_MANAGED_EXECUTION = False

    def __init__(
        self,
        log_stdout: str = "",
        log_stderr: str = "",
        cache_dir: str = "",
        time_limit: int = -1,
        retries: int = 0,
        *args,
        **kwargs,
    ):
        self.log_stdout = log_stdout
        self.log_stderr = log_stderr
        self.cache_dir = cache_dir
        self.time_limit = time_limit
        self.retries = retries

    def get_dispatch_context(self, dispatch_info: DispatchInfo) -> ContextManager[DispatchInfo]:
        """
        Start a context manager that will be used to
        access the dispatch info for the executor.

        Args:
            dispatch_info: The dispatch info to be used inside current context.

        Returns:
            A context manager object that handles the dispatch info.
        """

        return active_dispatch_info_manager.claim(dispatch_info)

    def short_name(self):
        return self.__module__.split("/")[-1].split(".")[-1]

    def to_dict(self) -> dict:
        """Return a JSON-serializable dictionary representation of self"""
        return {
            "type": str(self.__class__),
            "short_name": self.short_name(),
            "attributes": self.__dict__.copy(),
        }

    def from_dict(self, object_dict: dict) -> "BaseExecutor":
        """Rehydrate a dictionary representation

        Args:
            object_dict: a dictionary representation returned by `to_dict`

        Returns:
            self

        Instance attributes will be overwritten.
        """
        if object_dict:
            self.__dict__ = copy.deepcopy(object_dict["attributes"])
        return self

    @property
    def task_stdout(self):
        return self.__dict__.get("_task_stdout")

    @property
    def task_stderr(self):
        return self.__dict__.get("_task_stderr")


class BaseExecutor(_AbstractBaseExecutor):
    """
    Base executor class to be used for defining any executor
    plugin. Subclassing this class will allow you to define
    your own executor plugin which can be used in covalent.

    Attributes:
        log_stdout: The path to the file to be used for redirecting stdout.
        log_stderr: The path to the file to be used for redirecting stderr.
        cache_dir: The location used for cached files in the executor.
        time_limit: time limit for the task
        retries: Number of times to retry execution upon failure

    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

    def _init_runtime(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        cancel_pool: Optional[ThreadPoolExecutor] = None,
    ) -> None:
        """
        Create the required queues for cancel task messages to be shared back and forth

        Arg(s)
            loop: Asyncio event loop to create tasks on
            cancel_pool: A ThreadPoolExecutor object to submit tasks to

        Return(s)
            None
        """
        self._send_queue = asyncio.Queue()
        self._recv_queue = queue.Queue()
        self._loop = loop
        self._cancel_pool = cancel_pool

    def _notify(self, action: Signals, body: Any = None) -> None:
        """
        Notifies a waiting thread with the necessary action to take along with the arguments passed in as body

        Arg(s)
            action: One of three possible actions that a waiting thread must take -> Signals.GET, Signals.PUT, Signals.EXIT
            body: Respective arguments for each action

        Return(s)
            None
        """
        self._loop.call_soon_threadsafe(self._send_queue.put_nowait, (action, body))

    def _notify_sync(self, action: Signals, body: Any = None) -> Any:
        """
        Blocking version of the _notify method

        Arg(s)
            action: One of three possible actions that a waiting thread must take -> Signals.GET, Signals.PUT, Signals.EXIT
            body: Respective arguments for each action
        """
        self._notify(action, body)
        return self._wait_for_response()

    def _wait_for_response(self, timeout: int = 5) -> Any:
        """
        Wait for response from a thread depending on the action/body parameters sent

        Arg(s):
            timeout: Number of seconds to wait for the result to become available in the queue. Defaults to five seconds

        Return(s)
            body: Response to the corresponding action
        """
        status, body = self._recv_queue.get(timeout=timeout)
        if status is None:
            raise RuntimeError("Error waiting for response")
        return body

    def get_cancel_requested(self) -> bool:
        """
        Check if the task was requested to be cancelled by the user

        Arg(s)
            None

        Return(s)
            True/False whether task cancellation was requested
        """
        return self._notify_sync(Signals.GET, "cancel_requested")

    def get_version_info(self) -> Dict:
        """
        Query the database for the task's Python and Covalent version

        Arg:
            dispatch_id: Dispatch ID of the lattice

        Returns:
            {"python": python_version, "covalent": covalent_version}
        """
        return self._notify_sync(Signals.GET, "version_info")

    def set_job_handle(self, handle: TypeJSON) -> Any:
        """
        Save the job_id/handle returned by the backend executing the task

        Arg(s)
            handle: Any JSONable type to identifying the task being executed by the backend

        Return(s)
            Response from saving the job handle to database
        """
        return self._notify_sync(Signals.PUT, ("job_handle", json.dumps(handle)))

    def _set_job_status_sync(self, status: Status) -> bool:
        if self.validate_status(status):
            return self._notify_sync(Signals.PUT, ("job_status", str(status)))
        else:
            return False

    def validate_status(self, status: Status) -> bool:
        """Overridable filter"""
        return True

    async def set_job_status(self, status: Status) -> bool:
        """
        Sets the job state

        For use with send/receive API

        Return(s)
            Whether the action succeeded
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            self._set_job_status_sync,
            status,
        )

    def write_streams_to_file(
        self,
        stream_strings: Iterable[str],
        filepaths: Iterable[str],
        dispatch_id: str,
        results_dir: str,
    ) -> None:
        """
        Write the contents of stdout and stderr to respective files.

        Args:
            stream_strings: The stream_strings to be written to files.
            filepaths: The filepaths to be used for writing the streams.
            dispatch_id: The ID of the dispatch which initiated the request.
            results_dir: The location of the results directory.
        """

        for ss, filepath in zip(stream_strings, filepaths):
            if filepath:
                # If it is a relative path, attach to results dir
                if not Path(filepath).expanduser().is_absolute():
                    filepath = os.path.join(results_dir, dispatch_id, filepath)

                filename = Path(filepath)
                filename = filename.expanduser()
                filename.parent.mkdir(parents=True, exist_ok=True)
                filename.touch(exist_ok=True)

                with open(filepath, "a") as f:
                    f.write(ss)

    async def _execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            self.execute,
            function,
            args,
            kwargs,
            dispatch_id,
            results_dir,
            node_id,
        )

    def execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:
        """
        Execute the function with the given arguments.

        This calls the executor-specific `run()` method.

        Args:
            function: The input python function which will be executed and whose result
                      is ultimately returned by this function.
            args: List of positional arguments to be used by the function.
            kwargs: Dictionary of keyword arguments to be used by the function.
            dispatch_id: The unique identifier of the external lattice process which is
                         calling this function.
            results_dir: The location of the results directory.
            node_id: ID of the node in the transport graph which is using this executor.

        Returns:
            output: The result of the function execution.
        """

        DispatchInfo(dispatch_id)
        function.args[0].python_version
        self._task_stdout = io.StringIO()
        self._task_stderr = io.StringIO()

        task_metadata = {
            "dispatch_id": dispatch_id,
            "node_id": node_id,
            "results_dir": results_dir,
        }

        try:
            self.setup(task_metadata=task_metadata)
            result = self.run(function, args, kwargs, task_metadata)
            job_status = RESULT_STATUS.COMPLETED
        except TaskRuntimeError:
            job_status = RESULT_STATUS.FAILED
            result = None
        except TaskCancelledError:
            job_status = RESULT_STATUS.CANCELLED
            result = None
        finally:
            self._notify(Signals.EXIT)
            self.teardown(task_metadata=task_metadata)

        self.write_streams_to_file(
            (self._task_stdout.getvalue(), self._task_stderr.getvalue()),
            (self.log_stdout, self.log_stderr),
            dispatch_id,
            results_dir,
        )

        return (result, self._task_stdout.getvalue(), self._task_stderr.getvalue(), job_status)

    def setup(self, task_metadata: Dict) -> Any:
        """Placeholder to run any executor specific tasks"""
        pass

    @abstractmethod
    def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict) -> Any:
        """Abstract method to run a function in the executor.

        Args:
            function: The function to run in the executor
            args: List of positional arguments to be used by the function
            kwargs: Dictionary of keyword arguments to be used by the function.
            task_metadata: Dictionary of metadata for the task. Current keys are
                          `dispatch_id` and `node_id`

        Returns:
            output: The result of the function execution
        """

        raise NotImplementedError

    def cancel(self, task_metadata: Dict, job_handle: Any) -> bool:
        """
        Method to cancel the job identified uniquely by the `job_handle` (base class)

        Arg(s)
            task_metadata: Metadata of the task to be cancelled
            job_handle: Unique ID of the job assigned by the backend

        Return(s)
            False by default
        """
        app_log.debug(f"Cancel not implemented for executor {type(self)}")
        return False

    async def _cancel(self, task_metadata: Dict, job_handle: Any) -> bool:
        """
        Cancel the task in a non-blocking manner

        Arg(s)
            task_metadata: Metadata of the task to be cancelled
            job_handle: Unique ID of the job assigned by the backend

        Return(s)
           Result of the task cancellation
        """
        cancel_result = await self._loop.run_in_executor(
            self._cancel_pool, self.cancel, task_metadata, job_handle
        )
        await self._loop.run_in_executor(self._cancel_pool, self.teardown, task_metadata)
        return cancel_result

    def teardown(self, task_metadata: Dict) -> Any:
        """Placeholder to run any executor specific cleanup/teardown actions"""
        pass

    async def send(
        self,
        task_specs: List[Dict],
        resources: ResourceMap,
        task_group_metadata: Dict,
    ):
        """Submit a list of task references to the compute backend.

        Args:
            task_specs: a list of TaskSpecs
            resources: a ResourceMap mapping task assets to URIs
            task_group_metadata: a dictionary of metadata for the task group.
                                 Current keys are `dispatch_id`, `node_ids`,
                                 and `task_group_id`.

        The return value of `send()` will be passed directly into `poll()`.
        """
        # Schemas:
        #
        # Task spec:
        # {
        #     "function_id": int,
        #     "args_ids": List[int],
        #     "kwargs_ids": Dict[str, int],
        #     "deps_id": str,
        #     "call_before_id": str,
        #     "call_after_id": str,
        # }

        # resources:
        # {
        #     "functions": Dict[int, str],
        #     "inputs": Dict[int, str],
        #     "deps": Dict[str, str]
        # }

        # task_group_metadata:
        # {
        #     "dispatch_id": str,
        #     "node_ids": List[int],
        #     "task_group_id": int,
        # }

        # Assets are will be accessible by the compute backend
        # at the provided URIs

        # Covalent will upload all assets before invoking `send()`.

        raise NotImplementedError

    async def poll(self, task_group_metadata: Dict, data: Any):
        # To be run as a background task.  A callback will be
        # registered with the runner to invoke the receive()

        raise NotImplementedError

    async def receive(
        self,
        task_group_metadata: Dict,
        data: Any,
    ) -> List[TaskUpdate]:
        # Returns (output_uri, stdout_uri, stderr_uri,
        # exception_raised)

        # Job should have reached a terminal state by the time this is invoked.

        raise NotImplementedError

    def get_upload_uri(self, task_metadata: Dict, object_key: str):
        return ""


class AsyncBaseExecutor(_AbstractBaseExecutor):
    """Async base executor class to be used for defining any executor
    plugin. Subclassing this class will allow you to define
    your own executor plugin which can be used in covalent.

    This is analogous to `BaseExecutor` except the `run()` method,
    together with the optional `setup()` and `teardown()` methods, are
    coroutines.

    Attributes:
        log_stdout: The path to the file to be used for redirecting stdout.
        log_stderr: The path to the file to be used for redirecting stderr.
        cache_dir: The location used for cached files in the executor.
        time_limit: time limit for the task
        retries: Number of times to retry execution upon failure

    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

    def _init_runtime(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        cancel_pool: Optional[ThreadPoolExecutor] = None,
    ) -> None:
        """
        Initialize the send and receive queues for communication between dispatcher and the executor

        Arg(s)
            loop: Asyncio event loop
            cancel_pool: Instance of a threadpool executor class

        Return(s)
            None
        """
        self._send_queue = asyncio.Queue()
        self._recv_queue = asyncio.Queue()

    def _notify(self, action: Signals, body: Any = None) -> None:
        """
        Notify the listener with the corresponding signal (async)

        Arg(s)
            action: Signal to the listener to trigger the corresponding action
            body: Message to be sent to the listener

        Return(s)
            None
        """
        self._send_queue.put_nowait((action, body))

    async def _notify_sync(self, action: Signals, body: Any = None) -> Any:
        """
        Blocking call to the `_notify` method to wait for response

        Arg(s)
            action: Signal to the listener to trigger the corresponding action
            body: Message to be sent to the listener

        Return(s)
            Response from the listener
        """
        self._notify(action, body)
        return await self._wait_for_response()

    async def _wait_for_response(self, timeout: int = 5) -> Any:
        """
        Block the thread until a response is recevied

        Arg(s)
            timeout: Number of seconds to wait until timing out

        Return(s)
            Response from the listener
        """
        aw = self._recv_queue.get()
        status, body = await asyncio.wait_for(aw, timeout=timeout)
        if status is False:
            raise RuntimeError("Error waiting for response")
        return body

    async def get_cancel_requested(self) -> Any:
        """
        Get if the task was requested to be canceled

        Arg(s)
            None

        Return(s)
            Whether the task has been requested to be cancelled
        """
        return await self._notify_sync(Signals.GET, "cancel_requested")

    async def get_version_info(self) -> Dict:
        """
        Query the database for dispatch version metadata.

        Arg:
            dispatch_id: Dispatch ID of the lattice

        Returns:
            {"python": python_version, "covalent": covalent_version}
        """
        return await self._notify_sync(Signals.GET, "version_info")

    async def set_job_handle(self, handle: TypeJSON) -> Any:
        """
        Save the job handle to database

        Arg(s)
            handle: JSONable type identifying the job being executed by the backend

        Return(s)
            Response from the listener that handles inserting the job handle to database
        """
        return await self._notify_sync(Signals.PUT, ("job_handle", json.dumps(handle)))

    def validate_status(self, status: Status) -> bool:
        """Overridable filter"""
        return True

    async def set_job_status(self, status: Status) -> bool:
        """
        Validates and sets the job state

        For use with send/receive API

        Return(s)
            Whether the action succeeded
        """
        if self.validate_status(status):
            return await self._notify_sync(Signals.PUT, ("job_status", str(status)))
        else:
            return False

    async def write_streams_to_file(
        self,
        stream_strings: Iterable[str],
        filepaths: Iterable[str],
        dispatch_id: str,
        results_dir: str,
    ) -> None:
        """
        Write the contents of stdout and stderr to respective files.

        Args:
            stream_strings: The stream_strings to be written to files.
            filepaths: The filepaths to be used for writing the streams.
            dispatch_id: The ID of the dispatch which initiated the request.
            results_dir: The location of the results directory.

        This uses aiofiles to avoid blocking the event loop.
        """

        for ss, filepath in zip(stream_strings, filepaths):
            if filepath:
                # If it is a relative path, attach to results dir
                if not Path(filepath).expanduser().is_absolute():
                    filepath = os.path.join(results_dir, dispatch_id, filepath)

                filename = Path(filepath)
                filename = filename.expanduser()
                filename.parent.mkdir(parents=True, exist_ok=True)
                filename.touch(exist_ok=True)

                async with aiofiles.open(filepath, "a") as f:
                    await f.write(ss)

    async def _execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:
        return await self.execute(
            function,
            args,
            kwargs,
            dispatch_id,
            results_dir,
            node_id,
        )

    async def execute(
        self,
        function: Callable,
        args: List,
        kwargs: Dict,
        dispatch_id: str,
        results_dir: str,
        node_id: int = -1,
    ) -> Any:
        self._task_stdout = io.StringIO()
        self._task_stderr = io.StringIO()

        task_metadata = {
            "dispatch_id": dispatch_id,
            "node_id": node_id,
            "results_dir": results_dir,
        }

        try:
            await self.setup(task_metadata=task_metadata)
            result = await self.run(function, args, kwargs, task_metadata)
            job_status = RESULT_STATUS.COMPLETED
        except TaskCancelledError:
            job_status = RESULT_STATUS.CANCELLED
            result = None
        except TaskRuntimeError:
            job_status = RESULT_STATUS.FAILED
            result = None
        finally:
            self._notify(Signals.EXIT)
            await self.teardown(task_metadata=task_metadata)

        await self.write_streams_to_file(
            (self._task_stdout.getvalue(), self._task_stderr.getvalue()),
            (self.log_stdout, self.log_stderr),
            dispatch_id,
            results_dir,
        )

        return (result, self._task_stdout.getvalue(), self._task_stderr.getvalue(), job_status)

    async def setup(self, task_metadata: Dict):
        """Executor specific setup method"""
        pass

    async def teardown(self, task_metadata: Dict):
        """Executor specific teardown method"""
        pass

    @abstractmethod
    async def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict) -> Any:
        """Abstract method to run a function in the executor in async-aware manner.

        Args:
            function: The function to run in the executor
            args: List of positional arguments to be used by the function
            kwargs: Dictionary of keyword arguments to be used by the function.
            task_metadata: Dictionary of metadata for the task. Current keys are
                          `dispatch_id` and `node_id`

        Returns:
            output: The result of the function execution
        """

        raise NotImplementedError

    async def cancel(self, task_metadata: Dict, job_handle: Any) -> bool:
        """
        Method to cancel the job identified uniquely by the `job_handle` (base class)

        Arg(s)
            task_metadata: Metadata of the task to be cancelled
            job_handle: Unique ID of the job assigned by the backend

        Return(s)
            False by default
        """
        app_log.debug(f"Cancel not implemented for executor {type(self)}")
        return False

    async def _cancel(self, task_metadata: Dict, job_handle: Any) -> bool:
        """
        Cancel the task in a non-blocking manner

        Arg(s)
            task_metadata: Metadata of the task to be cancelled
            job_handle: Unique ID of the job assigned by the backend

        Return(s)
           Result of the task cancellation
        """
        return await self.cancel(task_metadata, job_handle)

    async def send(
        self,
        task_specs: List[TaskSpec],
        resources: ResourceMap,
        task_group_metadata: Dict,
    ) -> Any:
        """Submit a list of task references to the compute backend.

        Args:
            task_specs: a list of TaskSpecs
            resources: a ResourceMap mapping task assets to URIs
            task_group_metadata: A dictionary of metadata for the task group.
                                 Current keys are `dispatch_id`, `node_ids`,
                                 and `task_group_id`.

        The return value of `send()` will be passed directly into `poll()`.
        """
        # Schemas:
        #
        # Task spec:
        # {
        #     "function_id": int,
        #     "args_ids": List[int],
        #     "kwargs_ids": Dict[str, int],
        #     "deps_id": str,
        #     "call_before_id": str,
        #     "call_after_id": str,
        # }

        # resources:
        # {
        #     "functions": Dict[int, str],
        #     "inputs": Dict[int, str],
        #     "deps": Dict[str, str]
        # }

        # task_group_metadata:
        # {
        #     "dispatch_id": str,
        #     "node_ids": List[int],
        #     "task_group_id": int,
        # }

        # Assets are assumed to be accessible by the compute backend
        # at the provided URIs

        # Covalent will upload all assets before invoking send().

        raise NotImplementedError

    async def poll(self, task_group_metadata: Dict, data: Any) -> Any:
        """Block until the job has reached a terminal state.

        Args:
            task_group_metadata: A dictionary of metadata for the task group.
                                 Current keys are `dispatch_id`, `node_ids`,
                                 and `task_group_id`.
            data: The return value of send().

        The return value of `poll()` will be passed directly into receive().

        Raise `NotImplementedError` to indicate that the compute backend
        will notify the Covalent server asynchronously of job completion.

        """

        raise NotImplementedError

    async def receive(
        self,
        task_group_metadata: Dict,
        data: Any,
    ) -> List[TaskUpdate]:
        """Return a list of task updates.

        Each task must have reached a terminal state by the time this is invoked.

        Args:
            task_group_metadata: A dictionary of metadata for the task group.
                                 Current keys are `dispatch_id`, `node_ids`,
                                 and `task_group_id`.
            data: The return value of poll() or the request body of `/jobs/update`.

        Returns:
            Returns a list of task results, each a TaskUpdate dataclass
            of the form

            {
              "dispatch_id": dispatch_id,
              "node_id": node_id,
              "status": status,
              "assets": {
                  "output":  {
                    "remote_uri": output_uri,
                  },
                  "stdout":  {
                    "remote_uri": stdout_uri,
                  },
                  "stderr":  {
                    "remote_uri": stderr_uri,
                  },
                },
            }

            corresponding to the node ids (task_ids) specified in the
            `task_group_metadata`.  This might be a subset of the node
            ids in the originally submitted task group as jobs may
            notify Covalent asynchronously of completed tasks before
            the entire task group finishes running.

        """

        raise NotImplementedError

    def get_upload_uri(self, task_group_metadata: Dict, object_key: str):
        return ""



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time
from typing import Optional, Sequence, Union

import pennylane as qml
from local_sampler import QiskitLocalSampler
from pennylane.tape.qscript import QuantumScript
from pydantic import Field
from runtime_sampler import QiskitRuntimeSampler

from covalent._shared_files.config import get_config
from covalent._shared_files.qelectron_utils import import_from_path
from covalent.executor.qbase import (
    AsyncBaseQExecutor,
    BaseThreadPoolQExecutor,
    QCResult,
    get_asyncio_event_loop,
    get_thread_pool,
)

__all__ = [
    "IBMQExecutor",
    "QiskitExecutor",
]

_QEXECUTOR_PLUGIN_DEFAULTS = {
    "IBMQExecutor": {
        "backend": "ibmq_qasm_simulator",
        "ibmqx_token": "",
        "hub": "ibm-q",
        "group": "open",
        "project": "main",
    },
    "QiskitExecutor": {
        "device": "local_sampler",
        "backend": "ibmq_qasm_simulator",
        "ibmqx_token": "",
        "hub": "ibm-q",
        "group": "open",
        "project": "main",
        "options": {
            "optimization_level": 3,
            "resilience_level": 1,
        },
    },
}

_DEVICE_MAP = {
    "local_sampler": QiskitLocalSampler,
    "sampler": QiskitRuntimeSampler,
}


class IBMQExecutor(BaseThreadPoolQExecutor):
    """
    A quantum executor that uses the Pennylane native :code:`"qiskit.ibmq"` device to run
    circuits on IBM Quantum backends. The attributes :code:`backend`, :code:`ibmqx_token`,
    :code:`hub`, :code:`group`, and :code:`project` are taken from the Covalent
    configuration file by default, if available.

    Keyword Args:
        max_jobs: The maximum number of jobs that can be submitted to the backend
            concurrently. This number corresponds to the number of threads utilized
            by this executor. Defaults to 20.
       shots: The number of shots to use for the execution device. Overrides the
            :code:`shots` value from the original device if set to :code:`None` or
            a positive :code:`int`. The shots setting from the original device is
            is used by default.
        backend: The name of the IBM Quantum backend device. Defaults to
            :code:`"ibmq_qasm_simulator"`.
        ibmqx_token: The IBM Quantum API token.
        hub: An IBM Quantum hub name. Defaults to :code:`"ibm-q"`.
        group: An IBM Quantum group name. Defaults to :code:`"open"`.
        project: An IBM Quantum project name. Defaults to :code:`"main"`.

    """

    max_jobs: int = 20

    backend: str = Field(
        default_factory=lambda: get_config("qelectron")["IBMQExecutor"]["backend"]
    )
    ibmqx_token: str = Field(
        default_factory=lambda: get_config("qelectron")["IBMQExecutor"]["ibmqx_token"]
    )
    hub: str = Field(default_factory=lambda: get_config("qelectron")["IBMQExecutor"]["hub"])
    group: str = Field(default_factory=lambda: get_config("qelectron")["IBMQExecutor"]["group"])
    project: str = Field(
        default_factory=lambda: get_config("qelectron")["IBMQExecutor"]["project"]
    )

    def batch_submit(self, qscripts_list):
        p = get_thread_pool(self.max_jobs)
        jobs = []
        for qscript in qscripts_list:
            dev = qml.device(
                "qiskit.ibmq",
                wires=self.qelectron_info.device_wires,
                shots=self.override_shots,
                backend=self.backend,
                ibmqx_token=self.ibmqx_token,
                hub=self.hub,
                group=self.group,
                project=self.project,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )

            jobs.append(p.submit(self.run_circuit, qscript, dev, result_obj))

        return jobs


class QiskitExecutor(AsyncBaseQExecutor):
    """
    A quantum executor that lets the user run circuits on IBM Quantum backends,
    using runtime sessions and Qiskit primitives. The attributes :code:`device`,
    :code:`backend`, :code:`ibmqx_token`, :code:`hub`, :code:`group`, and
    :code:`project` are taken from the Covalent configuration file by default, if
    available.

    Keyword Args:
        device: The Qiskit primitive used to execute circuits. Valid values are
            :code:`"sampler"` and :code:`"local_sampler"`. The value :code:`"sampler"`
            corresponds to the Qiskit Runtime :code:`Sampler` primitive. The value
            :code:`"local_sampler"` corresponds to the Qiskit :code:`Sampler` primitive,
            which is entirely local.
        backend: The name of the IBM Quantum backend device. Defaults to
            :code:`"ibmq_qasm_simulator"`.
        ibmqx_token: The IBM Quantum API token.
        hub: An IBM Quantum hub name. Defaults to :code:`"ibm-q"`.
        group: An IBM Quantum group name. Defaults to :code:`"open"`.
        project: An IBM Quantum project name. Defaults to :code:`"main"`.
        shots: The number of shots to run per circuit. Defaults to 1024.
        single_job: Indicates whether or not all circuits are submitted
            to a single job or as separate jobs. Defaults to :code:`True`.
        max_time: An optional time limit for circuit execution on the IBM Quantum
            backend. Defaults to :code:`None`, i.e. no time limit.
        local_transpile: Indicates whether or not to transpile circuits before
            submitting to IBM Quantum. Defaults to :code:`False`.
        ibmqx_url: An optional URL for the Qiskit Runtime API.
        channel: An optional channel for the Qiskit Runtime API. Defaults to
            :code:`"ibm_quantum"`.
        instance: An alternate means to specify :code:`hub`, :code:`group`, and
            :code:`project`, formatted as :code:`"my-hub/my-group/my-project"`.
        cloud_instance: Same as :code:`instance` but for the case :code:`channel="ibm_cloud"`.
        options: A dictionary of options to pass to Qiskit Runtime. See
            :code:`qiskit_ibm_runtime.options.Options` for valid fields.
    """

    device: str = Field(
        default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["device"]
    )
    backend: str = Field(
        default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["backend"]
    )
    ibmqx_token: str = Field(
        default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["ibmqx_token"]
    )
    hub: str = Field(default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["hub"])
    group: str = Field(default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["group"])
    project: str = Field(
        default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["project"]
    )

    shots: Union[None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]] = 1024
    single_job: bool = False
    local_transpile: bool = False

    max_time: Optional[Union[int, str]] = None

    ibmqx_url: Optional[str] = None
    channel: str = "ibm_quantum"
    instance: str = ""
    cloud_instance: str = ""

    options: Optional[Union[dict, tuple]] = Field(
        # pylint: disable=unnecessary-lambda
        default_factory=lambda: get_config("qelectron")["QiskitExecutor"]["options"]
    )

    @property
    def device_init_kwargs(self):
        """
        Keyword arguments to pass to the device constructor.
        """
        return {
            "wires": self.qelectron_info.device_wires,
            "shots": self.qelectron_info.device_shots or self.override_shots,
            "backend_name": self.backend,
            "local_transpile": self.local_transpile,
            "max_time": self.max_time,
            "single_job": self.single_job,
            "options": self.options or {},
            "service_init_kwargs": {
                "ibmqx_token": self.ibmqx_token,
                "ibmqx_url": self.ibmqx_url,
                "channel": self.channel,
                "instance": self.instance,
                "cloud_instance": self.cloud_instance,
                "hub": self.hub,
                "group": self.group,
                "project": self.project,
            },
        }

    def execution_device(self) -> qml.QubitDevice:
        """
        Create a subclasses execution device that ensure correct output typing.
        """

        # Initialize a custom Pennylane device
        dev = _execution_device_factory(
            self.device,
            qnode_device_cls=import_from_path(self.qelectron_info.device_import_path),
            **self.device_init_kwargs,
        )

        # Set `pennylane.active_return()` status
        dev.pennylane_active_return = self.qelectron_info.pennylane_active_return
        return dev

    def batch_submit(self, qscripts_list):
        qscripts_list = list(qscripts_list)

        loop = get_asyncio_event_loop()
        tasks = []

        if self.single_job:
            # All QScripts are submitted as a single job

            # initialize a custom Pennylane device
            dev = self.execution_device()

            # initialize a result object
            result_obj = QCResult.with_metadata(device_name=dev.short_name, executor=self)

            # run qscripts asynchronously
            task = loop.create_task(self.run_all_circuits(qscripts_list, dev, result_obj))
            tasks.append(task)
        else:
            # Each QScript is submitted as a separate job
            for qscript in qscripts_list:
                # initialize a custom Pennylane device
                dev = self.execution_device()

                # initialize a result object
                result_obj = QCResult.with_metadata(device_name=dev.short_name, executor=self)

                # run qscripts asynchronously
                task = loop.create_task(self.run_circuit(qscript, dev, result_obj))
                tasks.append(task)

        return tasks

    async def run_circuit(
        self, tape, device, result_obj: QCResult
    ):  # pylint: disable=arguments-renamed
        """
        Allows a circuit to be submitted asynchronously.
        """

        start_time = time.perf_counter()
        results = qml.execute([tape], device, gradient_fn="best")

        await asyncio.sleep(0)

        results, metadatas = device.post_process(tape, results)
        end_time = time.perf_counter()

        result_obj.results = results
        result_obj.execution_time = end_time - start_time
        result_obj.metadata["execution_metadata"].extend(metadatas)

        return result_obj

    async def run_all_circuits(
        self, tapes, device, result_obj: QCResult
    ):  # pylint: disable=arguments-renamed
        """
        Allows multiple circuits to be submitted asynchronously into a single
        IBM Qiskit Runtime Job.
        """

        start_time = time.perf_counter()
        results = qml.execute(tapes, device, gradient_fn="best")

        await asyncio.sleep(0)

        results, metadatas = device.post_process_all(tapes, results)
        end_time = time.perf_counter()

        result_obj.results = results
        result_obj.execution_time = end_time - start_time
        result_obj.metadata["execution_metadata"].extend(metadatas)

        return result_obj

    def dict(self, *args, **kwargs):
        dict_ = super().dict(*args, **kwargs)
        # Needed to make the dict method hashable and JSON-able.
        dict_["options"] = tuple(dict_["options"].items())
        return dict_


def _execution_device_factory(device_name: str, qnode_device_cls, **kwargs):
    """
    Creates a subclassed Pennylane device to ensure correct output typing.
    """
    custom_device_cls = _DEVICE_MAP.get(device_name)
    if not custom_device_cls:
        raise ValueError(f"Unsupported Qiskit primitive device '{device_name}'.")

    class _QiskitExecutionDevice(custom_device_cls, qnode_device_cls):
        # pylint: disable=too-few-public-methods

        pennylane_active_return: bool = True

        """
        Wrapper that inherits from a Pennylane device class to extend the custom
        Pennylane-Qiskit device with any device-specific overridden methods.
        """

        @property
        def stopping_condition(self):
            """
            Needed to target :code:`support_operation` method of :code:`custom_device_cls`.

            NOTE: is identical to `pennylane._device.Device.stopping_condition`.
            """
            return qml.BooleanFn(
                lambda obj: not isinstance(obj, QuantumScript)
                and self.supports_operation(obj.name)
            )

    return _QiskitExecutionDevice(**kwargs)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil

from .. import File
from .transfer_strategy_base import FileTransferStrategy


class Shutil(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to copy files locally

    The copying is done in-process using shutil.copyfile.
    """

    def __init__(
        self,
    ):
        pass

    # return callable to copy files in the local file system
    def cp(self, from_file: File, to_file: File = File()) -> None:
        """
        Get a callable that copies a file from one location to another locally

        Args:
            from_file: File to copy from
            to_file: File to copy to. Defaults to File().

        Returns:
            A callable that copies a file from one location to another locally
        """

        def callable():
            shutil.copyfile(from_file.filepath, to_file.filepath)

        return callable

    # Local file operations only
    def upload(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError

    # Local file operations only
    def download(self, from_file: File, to_file: File) -> File:
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-backed lattice"""

from typing import Any, List

from sqlalchemy.orm import Session

from .._db import models
from .base import DispatchedObject
from .controller import Record
from .db_interfaces.lattice_utils import ASSET_KEYS  # nopycln: import
from .db_interfaces.lattice_utils import METADATA_KEYS  # nopycln: import
from .db_interfaces.lattice_utils import _meta_record_map, get_filters, set_filters
from .tg import ELECTRON_KEYS, _TransportGraph

LATTICE_KEYS = list(_meta_record_map.keys())


class LatticeMeta(Record[models.Lattice]):
    model = models.Lattice


class LatticeAsset(Record[models.LatticeAsset]):
    model = models.LatticeAsset


class Lattice(DispatchedObject[LatticeMeta, LatticeAsset]):
    meta_type = LatticeMeta
    asset_link_type = LatticeAsset

    metadata_keys = LATTICE_KEYS

    def __init__(
        self,
        session: Session,
        record: models.Lattice,
        bare: bool = False,
        *,
        keys: List = LATTICE_KEYS,
        electron_keys: List = ELECTRON_KEYS,
    ):
        self._id = record.id
        self._keys = keys
        fields = set(map(Lattice.meta_record_map, keys))

        self._metadata = LatticeMeta(session, record, fields=fields)
        self._assets = {}
        self._lattice_id = record.id

        self.transport_graph = _TransportGraph.get_compute_graph(
            session, self._lattice_id, bare, keys=electron_keys
        )

    @property
    def query_keys(self) -> List:
        return self._keys

    @property
    def metadata(self) -> LatticeMeta:
        return self._metadata

    @property
    def assets(self):
        return self._assets

    @classmethod
    def meta_record_map(cls: DispatchedObject, key: str) -> str:
        return _meta_record_map[key]

    def get_value(self, key: str, session: Session = None, refresh: bool = True):
        return get_filters[key](super().get_value(key, session, refresh))

    def set_value(self, key: str, val: Any, session: Session = None) -> None:
        super().set_value(key, set_filters[key](val), session)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a plugin executor module that is loaded at runtime by defining its default attributes in _EXECUTOR_PLUGIN_DEFAULTS
and optionally overriding the attributes in its constructor.

This module contains a RemoteExecutor class that provides abstract methods for implementing other remote executors.
Executor classes that inherit from the RemoteExecutor class can perform task execution by connecting with a remote environment
via SSH or Slurm or by interfacing with compute services in various cloud environments such as AWS, Azure, and GCP.
"""

import asyncio
from abc import abstractmethod
from typing import Any, Dict, Tuple

from covalent._shared_files import logger
from covalent.executor.base import AsyncBaseExecutor

# The plugin class name must be given by the executor_plugin_name attribute:
EXECUTOR_PLUGIN_NAME = "RemoteExecutor"

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    "poll_freq": 15,
    "remote_cache": ".cache/covalent",
    "credentials_file": "",
    "covalent_version": "",
}


class RemoteExecutor(AsyncBaseExecutor):
    """
    Async executor class that provides abstract methods for managing task execution on a remote machine.

    Attributes:
        poll_freq: Number of seconds to wait between polling for task status after a task has been submitted.
        remote_cache: Location where pickled inputs/outputs are stored.
        credentials_file: Location where credentials can be found.
    """

    def __init__(
        self,
        poll_freq: int = 15,
        remote_cache: str = "",
        credentials_file: str = "",
        covalent_version: str = "",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.poll_freq = poll_freq
        self.remote_cache = remote_cache
        self.credentials_file = credentials_file
        self.covalent_version = covalent_version

    @abstractmethod
    async def _validate_credentials(self) -> bool:
        """
        Abstract method to validate user credentials.

        Return:
            bool: True if the user credentials are valid and false if otherwise.
        """
        pass

    @abstractmethod
    async def _upload_task(self) -> None:
        """
        Abstract method that uploads the pickled function to the remote cache.
        """
        pass

    @abstractmethod
    async def submit_task(self, task_metadata: Dict) -> Any:
        """
        Abstract method that invokes the task on the remote backend.

        Args:
            task_metadata: Dictionary of metadata for the task. Current keys are
                          `dispatch_id` and `node_id`.
        Return:
            task_uuid: Task UUID defined on the remote backend.
        """
        pass

    @abstractmethod
    async def get_status(self) -> Any:
        """
        Abstract method that returns the status of an electron by querying the remote backend.
        """
        pass

    @abstractmethod
    async def _poll_task(self) -> Any:
        """
        Abstract method that polls the remote backend until the status of a workflow's execution
        is either COMPLETED or FAILED.
        """
        pass

    @abstractmethod
    async def query_result(self) -> Any:
        """
        Abstract method that retrieves the pickled result from the remote cache.
        """

    @abstractmethod
    async def cancel(self) -> bool:
        """
        Abstract method that sends a cancellation request to the remote backend.
        """
        pass

    @staticmethod
    async def run_async_subprocess(cmd) -> Tuple:
        """
        Invokes an async subprocess to run a command.
        """

        proc = await asyncio.create_subprocess_shell(
            cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        stdout, stderr = await proc.communicate()

        if stdout:
            app_log.debug(stdout)

        if stderr:
            app_log.debug(stderr)

        return proc, stdout, stderr



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to convert lattice -> LatticeSchema"""

from typing import Dict

from .._shared_files.schemas.lattice import (
    ASSET_FILENAME_MAP,
    AssetSchema,
    LatticeAssets,
    LatticeMetadata,
    LatticeSchema,
)
from .._workflow.lattice import Lattice
from .common import AssetType, load_asset, save_asset
from .transport_graph import deserialize_transport_graph, serialize_transport_graph

__all__ = [
    "serialize_lattice",
    "deserialize_lattice",
]


ASSET_TYPES = {
    "workflow_function": AssetType.TRANSPORTABLE,
    "workflow_function_string": AssetType.TEXT,
    "doc": AssetType.TEXT,
    "inputs": AssetType.TRANSPORTABLE,
    "named_args": AssetType.TRANSPORTABLE,
    "named_kwargs": AssetType.TRANSPORTABLE,
    "cova_imports": AssetType.JSONABLE,
    "lattice_imports": AssetType.TEXT,
    "hooks": AssetType.JSONABLE,
}


def _serialize_lattice_metadata(lat) -> LatticeMetadata:
    name = lat.__name__
    executor = lat.metadata["executor"]
    executor_data = lat.metadata["executor_data"]
    workflow_executor = lat.metadata["workflow_executor"]
    workflow_executor_data = lat.metadata["workflow_executor_data"]
    python_version = lat.python_version
    covalent_version = lat.covalent_version
    return LatticeMetadata(
        name=name,
        python_version=python_version,
        covalent_version=covalent_version,
        executor=executor,
        executor_data=executor_data,
        workflow_executor=workflow_executor,
        workflow_executor_data=workflow_executor_data,
    )


def _deserialize_lattice_metadata(meta: LatticeMetadata) -> dict:
    return {
        "__name__": meta.name,
        "python_version": meta.python_version,
        "covalent_version": meta.covalent_version,
        "metadata": {
            "executor": meta.executor,
            "executor_data": meta.executor_data,
            "workflow_executor": meta.workflow_executor,
            "workflow_executor_data": meta.workflow_executor_data,
        },
    }


def _serialize_lattice_assets(lat, storage_path: str) -> LatticeAssets:
    workflow_func_asset = save_asset(
        lat.workflow_function,
        ASSET_TYPES["workflow_function"],
        storage_path,
        ASSET_FILENAME_MAP["workflow_function"],
    )

    try:
        workflow_func_string = lat.workflow_function_string
    except AttributeError:
        workflow_func_string = ""
    workflow_func_str_asset = save_asset(
        workflow_func_string,
        ASSET_TYPES["workflow_function_string"],
        storage_path,
        ASSET_FILENAME_MAP["workflow_function_string"],
    )

    docstring = lat.__doc__
    docstring_asset = save_asset(
        docstring,
        ASSET_TYPES["doc"],
        storage_path,
        ASSET_FILENAME_MAP["doc"],
    )

    inputs_asset = save_asset(
        lat.inputs, ASSET_TYPES["inputs"], storage_path, ASSET_FILENAME_MAP["inputs"]
    )

    # Deprecate
    named_args_asset = save_asset(
        lat.named_args,
        ASSET_TYPES["named_args"],
        storage_path,
        ASSET_FILENAME_MAP["named_args"],
    )
    named_kwargs_asset = save_asset(
        lat.named_kwargs,
        ASSET_TYPES["named_kwargs"],
        storage_path,
        ASSET_FILENAME_MAP["named_kwargs"],
    )
    cova_imports_asset = save_asset(
        lat.cova_imports,
        ASSET_TYPES["cova_imports"],
        storage_path,
        ASSET_FILENAME_MAP["cova_imports"],
    )
    lattice_imports_asset = save_asset(
        lat.lattice_imports,
        ASSET_TYPES["lattice_imports"],
        storage_path,
        ASSET_FILENAME_MAP["lattice_imports"],
    )

    # NOTE: these are actually JSONable
    hooks_asset = save_asset(
        lat.metadata["hooks"],
        ASSET_TYPES["hooks"],
        storage_path,
        ASSET_FILENAME_MAP["hooks"],
    )

    return LatticeAssets(
        workflow_function=workflow_func_asset,
        workflow_function_string=workflow_func_str_asset,
        doc=docstring_asset,
        inputs=inputs_asset,
        named_args=named_args_asset,
        named_kwargs=named_kwargs_asset,
        cova_imports=cova_imports_asset,
        lattice_imports=lattice_imports_asset,
        hooks=hooks_asset,
    )


def _deserialize_lattice_assets(assets: LatticeAssets) -> dict:
    workflow_function = load_asset(assets.workflow_function, ASSET_TYPES["workflow_function"])
    workflow_function_string = load_asset(
        assets.workflow_function_string, ASSET_TYPES["workflow_function_string"]
    )
    doc = load_asset(assets.doc, ASSET_TYPES["doc"])
    inputs = load_asset(assets.inputs, ASSET_TYPES["inputs"])
    named_args = load_asset(assets.named_args, ASSET_TYPES["named_args"])
    named_kwargs = load_asset(assets.named_kwargs, ASSET_TYPES["named_kwargs"])
    cova_imports = load_asset(assets.cova_imports, ASSET_TYPES["cova_imports"])
    lattice_imports = load_asset(assets.lattice_imports, ASSET_TYPES["lattice_imports"])
    hooks = load_asset(assets.hooks, ASSET_TYPES["hooks"])
    return {
        "workflow_function": workflow_function,
        "workflow_function_string": workflow_function_string,
        "__doc__": doc,
        "inputs": inputs,
        "named_args": named_args,
        "named_kwargs": named_kwargs,
        "cova_imports": cova_imports,
        "lattice_imports": lattice_imports,
        "metadata": {
            "hooks": hooks,
        },
    }


def _get_lattice_custom_assets(lat: Lattice) -> Dict[str, AssetSchema]:
    if "custom_asset_keys" in lat.metadata:
        return {key: AssetSchema(size=0) for key in lat.metadata["custom_asset_keys"]}


def serialize_lattice(lat, storage_path: str) -> LatticeSchema:
    meta = _serialize_lattice_metadata(lat)
    assets = _serialize_lattice_assets(lat, storage_path)
    custom_assets = _get_lattice_custom_assets(lat)
    tg = serialize_transport_graph(lat.transport_graph, storage_path)

    return LatticeSchema(
        metadata=meta, assets=assets, custom_assets=custom_assets, transport_graph=tg
    )


def deserialize_lattice(model: LatticeSchema) -> Lattice:
    def dummy_function(x):
        return x

    lat = Lattice(dummy_function)

    attrs = _deserialize_lattice_metadata(model.metadata)
    assets = _deserialize_lattice_assets(model.assets)

    metadata = assets.pop("metadata")
    attrs.update(assets)
    attrs["metadata"].update(metadata)

    tg = deserialize_transport_graph(model.transport_graph)

    attrs["transport_graph"] = tg

    lat.__dict__.update(attrs)

    return lat



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Endpoints to update status of running tasks."""


from fastapi import APIRouter, Request

from covalent._shared_files import logger

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()


@router.put("/dispatches/{dispatch_id}/electrons/{node_id}/job")
async def update_task_status(dispatch_id: str, node_id: int, request: Request):
    """Updates the status of a running task.

    The request JSON will be passed to the task executor plugin's
    `receive()` method together with `dispatch_id` and `node_id`.
    """

    from .._core import runner_ng

    task_metadata = {
        "dispatch_id": dispatch_id,
        "node_id": node_id,
    }
    try:
        detail = await request.json()
        f"Task {task_metadata} marked ready with detail {detail}"
        # detail = {"status": Status(status.value.upper())}
        await runner_ng.mark_task_ready(task_metadata, detail)
        # app_log.debug(f"Marked task {dispatch_id}:{node_id} with status {status}")
        return f"Task {task_metadata} marked ready with detail {detail}"
    except Exception as e:
        app_log.debug(f"Exception in update_task_status: {e}")
        raise



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for server-side analogues of workflow data types"""

from abc import abstractmethod
from typing import Any, Dict, Generator, Generic, List, Type, TypeVar, Union

from sqlalchemy import select
from sqlalchemy.orm import Session, load_only

from .._db.datastore import workflow_db
from . import controller
from .asset import FIELDS, Asset

# Metadata
MetaType = TypeVar("MetaType", bound=controller.Record)

# Asset links
AssetLinkType = TypeVar("AssetLinkType", bound=controller.Record)


class DispatchedObject(Generic[MetaType, AssetLinkType]):
    """Base class for types with both metadata and assets.

    Each subclass must define two properties:
    - `meta_type`: The controller class for the type's "pure metadata" table
    - `asset_link_type`: The controller class for the type's asset-links table

    """

    @classmethod
    @property
    def meta_type(cls) -> Type[MetaType]:
        """Returns the metadata controller class."""
        raise NotImplementedError

    @classmethod
    @property
    def asset_link_type(cls) -> Type[AssetLinkType]:
        """Returns the asset link controller class"""
        raise NotImplementedError

    @classmethod
    @property
    def metadata_keys(cls) -> set:
        raise NotImplementedError

    @property
    @abstractmethod
    def query_keys(self) -> set:
        raise NotImplementedError

    @property
    @abstractmethod
    def metadata(self) -> MetaType:
        raise NotImplementedError

    @property
    def computed_fields(self) -> Dict:
        return {}

    @classmethod
    def session(cls) -> Generator[Session, None, None]:
        return workflow_db.session()

    def get_asset_ids(self, session: Session, keys: List[str]) -> Dict[str, int]:
        membership_filters = {"key": keys} if len(keys) > 0 else {}
        records = type(self).asset_link_type.get(
            session,
            fields=[],
            equality_filters={"meta_id": self._id},
            membership_filters=membership_filters,
        )
        return {x.key: x.asset_id for x in records}

    def associate_asset(
        self, session: Session, key: str, asset_id: int, flush: bool = False
    ) -> AssetLinkType:
        asset_link_kwargs = {
            "meta_id": self._id,
            "asset_id": asset_id,
            "key": key,
        }
        return type(self).asset_link_type.create(
            session, insert_kwargs=asset_link_kwargs, flush=flush
        )

    @property
    @abstractmethod
    def assets(self) -> Dict[str, Asset]:
        raise NotImplementedError

    @classmethod
    def meta_record_map(cls, key: str) -> str:
        return key

    def _refresh_metadata(self, session: Session, *, for_update: bool = False):
        fields = {type(self).meta_record_map(k) for k in self.query_keys}
        self.metadata.refresh(session, fields=fields, for_update=for_update)

    def get_metadata(self, key: str, session: Session, refresh: bool = True):
        attr = type(self).meta_record_map(key)
        if refresh:
            self._refresh_metadata(session)
        return self.metadata.attrs[attr]

    def set_metadata(self, key: str, val: Union[str, int], session: Session):
        record_attr = type(self).meta_record_map(key)
        self.metadata.update(session, values={record_attr: val})

    def incr_metadata(self, key: str, delta: int, session: Session):
        attr = type(self).meta_record_map(key)
        self.metadata.incr(session, increments={attr: delta})

    def get_asset(self, key: str, session: Session, refresh: bool = True) -> Asset:
        if key in self.assets:
            if refresh:
                self.assets[key].refresh(session, fields=FIELDS)
        else:
            asset_id = self.get_asset_ids(session, [key])[key]
            self.assets[key] = Asset.from_id(asset_id, session)

        return self.assets[key]

    def populate_asset_map(self, session: Session):
        """Load and cache all asset records"""
        asset_links = self.get_asset_ids(session=session, keys=[])
        for key, asset_id in asset_links.items():
            self.assets[key] = Asset.from_id(asset_id, session)

    def update_assets(self, updates: Dict[str, Dict], session: Session):
        """Bulk update associated assets"""
        for key, values in updates.items():
            asset = self.get_asset(key, session)
            asset.update(session, values=values)

    def _get_value(self, key: str, session: Session, refresh: bool = True) -> Any:
        if key in self.computed_fields:
            handler = self.computed_fields[key]
            return handler(self, session)
        elif key in self.metadata_keys:
            return self.get_metadata(key, session, refresh)
        else:
            return self.get_asset(key, session).load_data()

    def get_value(self, key: str, session: Session = None, refresh: bool = True) -> Any:
        if session is not None:
            return self._get_value(key, session, refresh)
        else:
            with self.session() as session:
                return self._get_value(key, session, refresh)

    def _set_value(self, key: str, val: Any, session: Session) -> None:
        if key in type(self).metadata_keys:
            self.set_metadata(key, val, session)
        else:
            self.get_asset(key, session).store_data(val, session)

    def set_value(self, key: str, val: Any, session: Session = None) -> None:
        if session is not None:
            self._set_value(key, val, session)
        else:
            with self.session() as session:
                self._set_value(key, val, session)

    def get_values(self, keys: List[str], session: Session = None, refresh: bool = True) -> Dict:
        return {key: self.get_value(key, session, refresh) for key in keys}

    @classmethod
    def get_db_records(
        cls,
        session: Session,
        *,
        keys: list,
        equality_filters: dict,
        membership_filters: dict,
        for_update: bool = False,
    ):
        # transform keys to db field names
        fields = list(map(cls.meta_record_map, keys))

        eq_filters_transformed = {}
        member_filters_transformed = {}
        for key, val in equality_filters.items():
            attr = cls.meta_record_map(key)
            eq_filters_transformed[attr] = val
        for key, vals in membership_filters.items():
            attr = cls.meta_record_map(key)
            member_filters_transformed[attr] = vals

        return cls.meta_type.get(
            session,
            fields=fields,
            equality_filters=eq_filters_transformed,
            membership_filters=member_filters_transformed,
            for_update=for_update,
        )

    @classmethod
    def get_linked_assets(
        cls, session, *, fields: list, equality_filters: dict, membership_filters: dict
    ) -> List[Asset]:
        link_model = cls.asset_link_type.model
        stmt = (
            select(link_model.meta_id, link_model.key, Asset.model)
            .join(link_model)
            .join(cls.meta_type.model)
        )
        if len(fields) == 0:
            fields = FIELDS
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.meta_type.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.meta_type.model, attr).in_(vals))

        attrs = [getattr(Asset.model, f) for f in fields]
        stmt = stmt.options(load_only(*attrs))

        records = session.execute(stmt)

        return [
            {"meta_id": row.meta_id, "key": row.key, "asset": Asset(session, row[2], keys=fields)}
            for row in records
        ]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Monitor executor instances """


from typing import Any

from covalent._shared_files import logger
from covalent.executor.base import _AbstractBaseExecutor as _ABE
from covalent.executor.utils import Signals

from .jobs import (
    get_cancel_requested,
    get_job_status,
    get_version_info,
    put_job_handle,
    put_job_status,
)

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_putters = {}
_getters = {}


_putters["job_handle"] = put_job_handle
_putters["job_status"] = put_job_status
_getters["cancel_requested"] = get_cancel_requested
_getters["job_status"] = get_job_status
_getters["version_info"] = get_version_info


async def _handle_message(
    dispatch_id: str, task_id: int, executor: _ABE, action: Signals, body: Any = None
) -> Any:
    """
    Handle the action properly

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task in the lattice
        executor: Instance of the abstract base executor
        action: Action requested to be performed
        body: Content of the action/request made

    Return(s)
        Response corresponding to the action requested
    """
    if action == Signals.GET:
        return await _getters[body](dispatch_id, task_id)

    if action == Signals.PUT:
        key, val = body
        await _putters[key](dispatch_id, task_id, val)
        return None
    else:
        raise KeyError(f"Unknown action {action}")


async def watch(dispatch_id: str, task_id: int, executor: _ABE) -> None:
    """
    Watch the send and receive queues of the executor

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task within the lattice
        executor: Instance of the abstract base executor

    Return(s)
        None
    """
    send_queue = executor._send_queue
    recv_queue = executor._recv_queue

    while True:
        action, body = await send_queue.get()
        app_log.debug(f"Received message {action} from executor {dispatch_id}:{task_id}")

        if action == Signals.EXIT:
            app_log.debug(f"Stopping listener for executor {dispatch_id}:{task_id}")
            break

        try:
            resp = await _handle_message(dispatch_id, task_id, executor, action, body)
            recv_queue.put_nowait((True, resp))
        except Exception as ex:
            app_log.warning(f"Error handling message {action} from executor: {ex}")
            recv_queue.put_nowait((False, None))
            break



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-backed lattice"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List

from sqlalchemy.orm import Session

from covalent._shared_files import logger
from covalent._shared_files.defaults import postprocess_prefix
from covalent._shared_files.util_classes import RESULT_STATUS, Status

from .._db import models
from .asset import Asset, copy_asset, copy_asset_meta
from .base import DispatchedObject
from .controller import Record
from .db_interfaces.result_utils import ASSET_KEYS  # nopycln: import
from .db_interfaces.result_utils import METADATA_KEYS  # nopycln: import
from .db_interfaces.result_utils import _meta_record_map, get_filters, set_filters
from .electron import ELECTRON_KEYS, Electron
from .lattice import LATTICE_KEYS, Lattice

app_log = logger.app_log

RESULT_KEYS = list(_meta_record_map.keys())


class ResultMeta(Record[models.Lattice]):
    model = models.Lattice


class ResultAsset(Record[models.LatticeAsset]):
    model = models.LatticeAsset


class Result(DispatchedObject[ResultMeta, ResultAsset]):
    meta_type = ResultMeta
    asset_link_type = ResultAsset
    metadata_keys = RESULT_KEYS

    def __init__(
        self,
        session: Session,
        record: models.Lattice,
        bare: bool = False,
        *,
        keys: list = RESULT_KEYS,
        lattice_keys: list = LATTICE_KEYS,
        electron_keys: list = ELECTRON_KEYS,
    ):
        self._id = record.id
        self._keys = keys
        fields = set(map(Result.meta_record_map, keys))
        self._metadata = ResultMeta(session, record, fields=fields)
        self._assets = {}

        self._lattice_id = record.id
        self._electron_id = record.electron_id

        self.lattice = Lattice(
            session, record, bare, keys=lattice_keys, electron_keys=electron_keys
        )

        self._task_failed = False
        self._task_cancelled = False

        # For lattice updates
        self._start_time = None
        self._end_time = None
        self._status = None
        self._error = None
        self._result = None

    @property
    def query_keys(self) -> List:
        return self._keys

    @property
    def metadata(self) -> ResultMeta:
        return self._metadata

    @property
    def assets(self):
        return self._assets

    @classmethod
    def meta_record_map(cls: DispatchedObject, key: str) -> str:
        return _meta_record_map[key]

    @property
    def start_time(self):
        return self.get_value("start_time")

    @property
    def end_time(self):
        return self.get_value("end_time")

    @property
    def dispatch_id(self):
        return self.get_value("dispatch_id")

    @property
    def root_dispatch_id(self):
        return self.get_value("root_dispatch_id")

    @property
    def status(self) -> Status:
        return self.get_value("status")

    @property
    def result(self):
        return self.get_value("result")

    @property
    def error(self):
        return self.get_value("error")

    def get_value(self, key: str, session: Session = None, refresh: bool = True):
        return get_filters[key](super().get_value(key, session, refresh))

    def set_value(self, key: str, val: Any, session: Session = None) -> None:
        super().set_value(key, set_filters[key](val), session)

    def _update_dispatch(
        self,
        start_time: datetime = None,
        end_time: datetime = None,
        status: "Status" = None,
        error: str = None,
        result: Any = None,
    ):
        """
        Update the dispatch metadata.

        Args:
            start_time: The start time of the lattice execution.
            end_time: The end time of the lattice execution.
            status: The status of the lattice execution.
            result: The lattice output unless error occured in which case None.
            error: Any error that occurred

        """

        with self.session() as session:
            if start_time is not None:
                self.set_value("start_time", start_time, session)
            if end_time is not None:
                self.set_value("end_time", end_time, session)
            if status is not None:
                self.set_value("status", status, session)
            if error is not None:
                self.set_value("error", error, session)
            if result is not None:
                self.set_value("result", result, session)

        # Copy output and error assets to sublattice's parent electron
        if RESULT_STATUS.is_terminal(status) and self._electron_id:
            with self.session() as session:
                electron_rec = Electron.get_db_records(
                    session,
                    keys={"id", "parent_lattice_id"},
                    equality_filters={"id": self._electron_id},
                    membership_filters={},
                )[0]
                parent_electron = Electron(session, electron_rec)

                subl_output = self.get_asset("result", session)
                subl_err = self.get_asset("error", session)
                electron_output = parent_electron.get_asset("output", session)
                electron_err = parent_electron.get_asset("error", session)

            app_log.debug("Copying sublattice output to parent electron")
            with self.session() as session:
                copy_asset_meta(session, subl_output, electron_output)
                copy_asset_meta(session, subl_err, electron_err)

            copy_asset(subl_output, electron_output)
            copy_asset(subl_err, electron_err)

    def _update_node(
        self,
        node_id: int,
        node_name: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        status: "Status" = None,
        output: Any = None,
        error: Exception = None,
        stdout: str = None,
        stderr: str = None,
        qelectron_data_exists: bool = None,
    ) -> bool:
        """
        Update the node result in the transport graph.
        Called after any change in node's execution state.

        Args:
            node_id: The node id.
            node_name: The name of the node.
            start_time: The start time of the node execution.
            end_time: The end time of the node execution.
            status: The status of the node execution.
            output: The output of the node unless error occured in which case None.
            error: The error of the node if occured else None.
            stdout: The stdout of the node execution.
            stderr: The stderr of the node execution.
            qelectron_data_exists: Whether the qelectron data exists.

        Returns:
            True/False indicating whether the update succeeded
        """

        app_log.debug("Inside update node")

        _start_ts = datetime.now()
        with self.session() as session:
            if status is not None:
                # This acquires a lock on the electron's row to achieve atomic RMW
                if self._can_update_node_status(session, node_id, status):
                    self.lattice.transport_graph.set_node_value(node_id, "status", status, session)
                    if status == RESULT_STATUS.COMPLETED:
                        self.incr_metadata("completed_electron_num", 1, session)
                else:
                    # Abort the update if illegal status update
                    session.rollback()
                    return False

            # Current node name
            name = self.lattice.transport_graph.get_node_value(node_id, "name", session)

            if node_name is not None:
                self.lattice.transport_graph.set_node_value(node_id, "name", node_name, session)

            if start_time is not None:
                self.lattice.transport_graph.set_node_value(
                    node_id, "start_time", start_time, session
                )

            if end_time is not None:
                self.lattice.transport_graph.set_node_value(node_id, "end_time", end_time, session)

            if output is not None:
                self.lattice.transport_graph.set_node_value(node_id, "output", output, session)

            if error is not None:
                self.lattice.transport_graph.set_node_value(node_id, "error", error, session)

            if stdout is not None:
                self.lattice.transport_graph.set_node_value(node_id, "stdout", stdout, session)

            if stderr is not None:
                self.lattice.transport_graph.set_node_value(node_id, "stderr", stderr, session)

            if qelectron_data_exists is not None:
                self.lattice.transport_graph.set_node_value(
                    node_id, "qelectron_data_exists", qelectron_data_exists, session
                )

        # Handle postprocessing node
        tg = self.lattice.transport_graph
        if name.startswith(postprocess_prefix) and end_time is not None:
            app_log.debug(f"Postprocess status: {status}")
            # Copy asset metadata
            with self.session() as session:
                workflow_result = self.get_asset("result", session)
                node_output = tg.get_node(node_id).get_asset("output", session)
                copy_asset_meta(session, node_output, workflow_result)
            copy_asset(node_output, workflow_result)

            self._update_dispatch(status=status, end_time=end_time)

        _end_ts = datetime.now()
        dt = (_end_ts - _start_ts).total_seconds()
        app_log.debug(f"_update_node took {dt} seconds")
        return True

    def _can_update_node_status(self, session: Session, node_id: int, new_status: Status) -> bool:
        """Checks whether a node status update is valid.

        The following status transitions are disallowed:
        * same-status updates e.g. completed -> completed
        * transitions from a terminal status

        In addition, a terminal status update for a sublattice electron must be consistent with the sublattice dispatch's status.

        Returns:
            bool: Whether the status update is allowed

        Side effects:
            This uses SELECT FOR UPDATE to acquire a row lock
        """

        node = self.lattice.transport_graph.get_node(node_id, session)
        node._refresh_metadata(session, for_update=True)
        old_status = node.get_value("status", session, refresh=False)
        if RESULT_STATUS.is_terminal(old_status) or old_status == new_status:
            app_log.debug(
                f"{self.dispatch_id}:{node_id}: illegal status update {old_status} -> {new_status}"
            )
            return False

        # If node is a sublattice electron, ensure that terminal
        # status updates agree with the sublattice dispatch status
        node_type = node.get_value("type", session, refresh=False)
        if node_type == "sublattice" and RESULT_STATUS.is_terminal(new_status):
            # Fetch sublattice result
            sub_dispatch_id = node.get_value("sub_dispatch_id", session, refresh=False)
            if sub_dispatch_id:
                sub_result = Result.from_dispatch_id(sub_dispatch_id, bare=True)
                if sub_result.status != new_status:
                    return False

        return True

    def _get_failed_nodes(self) -> List[int]:
        """
        Get the node_id of each failed task
        """
        return self._get_incomplete_nodes()["failed"]

    def _get_incomplete_nodes(self):
        """
        Get all nodes that did not complete.

        Returns:
            A dictionary {"failed": [node_ids], "cancelled": [node_ids]}
        """
        with self.session() as session:
            query_keys = {"parent_lattice_id", "node_id", "name", "status"}
            records = Electron.get_db_records(
                session,
                keys=query_keys,
                equality_filters={"parent_lattice_id": self._id},
                membership_filters={
                    "status": [str(RESULT_STATUS.FAILED), str(RESULT_STATUS.CANCELLED)]
                },
            )

            nodes = list(map(lambda rec: Electron(session, rec, keys=query_keys), records))

            failed = list(filter(lambda e: e.get_value("status") == RESULT_STATUS.FAILED, nodes))
            cancelled = list(
                filter(lambda e: e.get_value("status") == RESULT_STATUS.CANCELLED, nodes)
            )

            failed_nodes = list(
                map(lambda x: (x.node_id, x.get_metadata("name", session, False)), failed)
            )
            cancelled_nodes = list(
                map(lambda x: (x.node_id, x.get_metadata("name", session, False)), cancelled)
            )

        return {"failed": failed_nodes, "cancelled": cancelled_nodes}

    def get_all_node_outputs(self) -> dict:
        """
        Return output of every node execution.

        Args:
            None

        Returns:
            node_outputs: A dictionary containing the output of every node execution.
        """

        all_node_outputs = {}
        tg = self.lattice.transport_graph
        for node_id in tg._graph.nodes:
            node_name = tg.get_node_value(node_id, "name")
            node_output = tg.get_node_value(node_id, "output")
            all_node_outputs[f"{node_name}({node_id})"] = node_output
        return all_node_outputs

    def get_all_assets(self, include_nodes: bool = True) -> Dict[str, List[Asset]]:
        assets = {}

        with self.session() as session:
            lattice_records = type(self).get_linked_assets(
                session,
                fields=[],
                equality_filters={"id": self._id},
                membership_filters={},
            )
            assets["lattice"] = list(map(lambda r: r["asset"], lattice_records))
            if include_nodes:
                node_records = Electron.get_linked_assets(
                    session,
                    fields=[],
                    equality_filters={"parent_lattice_id": self._lattice_id},
                    membership_filters={},
                )
                assets["nodes"] = list(map(lambda r: r["asset"], node_records))
        return assets

    @classmethod
    def ensure_run_once(cls, dispatch_id: str) -> bool:
        """Ensure that a dispatch is only run once.


        Returns:
            bool: whether the dispatch can be run
        """
        # Atomically increment dispatch status from NEW_OBJ to STARTING
        with cls.session() as session:
            record = ResultMeta.get(
                session,
                fields=["id", "dispatch_id", "status"],
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
                for_update=True,
            )[0]
            status = get_filters["status"](record.status)

            if status == RESULT_STATUS.NEW_OBJECT:
                new_status = set_filters["status"](RESULT_STATUS.STARTING)
                ResultMeta.update_bulk(
                    session,
                    values={"status": new_status},
                    equality_filters={"dispatch_id": dispatch_id},
                    membership_filters={},
                )
                app_log.debug(f"dispatch {dispatch_id} has not been run")
                return True
            else:
                app_log.debug(f"dispatch {dispatch_id} has already been run")
                return False

    @classmethod
    def from_dispatch_id(
        cls,
        dispatch_id: str,
        bare: bool,
        *,
        session: Session = None,
        keys: list = RESULT_KEYS,
        lattice_keys: list = LATTICE_KEYS,
        electron_keys: list = ELECTRON_KEYS,
    ) -> Result:
        if session:
            records = Result.get_db_records(
                session,
                keys=keys + lattice_keys,
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            if not records:
                raise KeyError(f"Dispatch {dispatch_id} not found")

            record = records[0]

            return Result(
                session,
                record,
                bare,
                keys=keys,
                lattice_keys=lattice_keys,
                electron_keys=electron_keys,
            )
        else:
            _start_ts = datetime.now()
            with Result.session() as session:
                _lock_ts = datetime.now()
                _lock_dt = (_lock_ts - _start_ts).total_seconds()
                app_log.debug(f"Acquiring db session took {_lock_dt} seconds")
                records = Result.get_db_records(
                    session,
                    keys=keys + lattice_keys,
                    equality_filters={"dispatch_id": dispatch_id},
                    membership_filters={},
                )
                if not records:
                    raise KeyError(f"Dispatch {dispatch_id} not found")

                record = records[0]

                res = Result(
                    session,
                    record,
                    bare,
                    keys=keys,
                    lattice_keys=lattice_keys,
                    electron_keys=electron_keys,
                )

                _end_ts = datetime.now()

                dt = (_end_ts - _start_ts).total_seconds()
                app_log.debug(f"get_result_object (bare={bare}) took {dt} seconds")
                return res


def get_result_object(
    dispatch_id: str,
    bare: bool = True,
    *,
    session: Session = None,
    keys: list = RESULT_KEYS,
    lattice_keys: list = LATTICE_KEYS,
    electron_keys: list = ELECTRON_KEYS,
) -> Result:
    return Result.from_dispatch_id(
        dispatch_id,
        bare,
        session=session,
        keys=keys,
        lattice_keys=lattice_keys,
        electron_keys=electron_keys,
    )
